from plone.namedfile import field
from plone.namedfile import file
from plone.namedfile.testing import PLONE_NAMEDFILE_INTEGRATION_TESTING
from plone.namedfile.tests import getFile

import unittest


class TestValidation(unittest.TestCase):

    layer = PLONE_NAMEDFILE_INTEGRATION_TESTING

    def test_validation_NamedImage_default(self):
        # Testing the default accepted media types
        image_field = field.NamedImage(
            required=False,
        )

        # field is empty, passes
        image_field.validate(None)

        # field has an empty file, fails
        # NOTE: This fails not because the NamedFile is empty but because the
        #       fallback default mimetype is "application/octet-stream".
        #       Not sure, if we should change this behavior.
        #       See: plone.namedfile.utils.get_contenttype
        named_image = file.NamedImage()
        self.assertRaises(field.InvalidImageFile, image_field.validate, named_image)

        # field has an png image file, passes
        named_image = file.NamedImage(getFile("image.png"), filename="image.png")
        image_field.validate(named_image)

        # field has an gif image file, passes
        named_image = file.NamedImage(getFile("image.gif"), filename="image.gif")
        image_field.validate(named_image)

        # field has a non-image file, fails
        named_image = file.NamedImage(getFile("notimage.doc"), filename="notimage.doc")
        self.assertRaises(field.InvalidImageFile, image_field.validate, named_image)

    def test_validation_NamedImage_custom(self):
        # Testing the default accepted media types
        image_field = field.NamedImage(
            accept=("image/png", ".jpg"),
            required=False,
        )

        # field is empty, passes
        image_field.validate(None)

        # field has an empty file, fails
        # NOTE: This fails not because the NamedFile is empty but because the
        #       fallback default mimetype is "application/octet-stream".
        #       Not sure, if we should change this behavior.
        #       See: plone.namedfile.utils.get_contenttype
        named_image = file.NamedImage()
        self.assertRaises(field.InvalidImageFile, image_field.validate, named_image)

        # field has a png image file, passes
        named_image = file.NamedImage(getFile("image.png"), filename="image.png")
        image_field.validate(named_image)

        # field has a jpg image file, passes also
        named_image = file.NamedImage(getFile("image.jpg"), filename="image.jpg")
        image_field.validate(named_image)

        # field has a gif image file, fails because it's not in the accepted
        # media types
        named_image = file.NamedImage(getFile("image.gif"), filename="image.gif")
        self.assertRaises(field.InvalidImageFile, image_field.validate, named_image)

        # field has a non-image file, fails
        named_image = file.NamedImage(getFile("notimage.doc"), filename="notimage.doc")
        self.assertRaises(field.InvalidImageFile, image_field.validate, named_image)

    def test_validation_NamedFile_default(self):
        # Testing the default accepted media types
        file_field = field.NamedFile(
            required=False,
        )

        # field is empty, passes
        file_field.validate(None)

        # field has n pdf file file, passes
        named_file = file.NamedFile(getFile("file.pdf"), filename="file.pdf")
        file_field.validate(named_file)

        # field has a gif file, passes
        named_file = file.NamedFile(getFile("image.gif"), filename="image.gif")
        file_field.validate(named_file)

    def test_validation_NamedFile_custom(self):
        # Testing the default accepted media types
        file_field = field.NamedFile(
            accept=("application/pdf", ".jpg"),
            required=False,
        )

        # field is empty, passes
        file_field.validate(None)

        # field has a pdf file file, passes
        named_file = file.NamedFile(getFile("file.pdf"), filename="file.pdf")
        file_field.validate(named_file)

        # field has a jpg file file, passes also
        named_file = file.NamedFile(getFile("image.jpg"), filename="image.jpg")
        file_field.validate(named_file)

        # field has a gif file, fails because it's not in the accepted media
        # types
        named_file = file.NamedFile(getFile("image.gif"), filename="image.gif")
        self.assertRaises(field.InvalidFile, file_field.validate, named_file)



from plone.dexterity.interfaces import IDexterityContent
from plone.namedfile.interfaces import INamedImageField
from plone.registry.interfaces import IRegistry
from zope.component import adapter
from zope.component import getMultiAdapter
from zope.component import getUtility
from zope.interface import implementer
from zope.interface import Interface
from zope.interface.interfaces import ComponentLookupError


try:
    from plone.base.interfaces import IImageScalesFieldAdapter
    from plone.base.interfaces import IImagingSchema
except ImportError:
    # BBB Plone 5
    IImageScalesFieldAdapter = Interface
    IImagingSchema = None


def _split_scale_info(allowed_size):
    name, dims = allowed_size.split(" ")
    width, height = list(map(int, dims.split(":")))
    return name, width, height


def _get_scale_infos():
    """Returns list of (name, width, height) of the available image scales."""
    if IImagingSchema is None:
        return []
    registry = getUtility(IRegistry)
    imaging_settings = registry.forInterface(IImagingSchema, prefix="plone")
    allowed_sizes = imaging_settings.allowed_sizes
    return [_split_scale_info(size) for size in allowed_sizes]


@implementer(IImageScalesFieldAdapter)
@adapter(INamedImageField, IDexterityContent, Interface)
class ImageFieldScales:
    def __init__(self, field, context, request):
        self.context = context
        self.request = request
        self.field = field

    def __call__(self):
        image = self.field.get(self.context)
        if not image:
            return

        # Get the @@images view once and store it, so all methods can use it.
        try:
            self.images_view = getMultiAdapter((self.context, self.request), name="images")
        except ComponentLookupError:
            # Seen in plone.app.caching.tests.test_profile_with_caching_proxy.
            # If we cannot find the images view, there is nothing for us to do.
            return
        width, height = image.getImageSize()
        url = self.get_original_image_url(self.field.__name__, width, height)
        scales = self.get_scales(self.field, width, height)

        # Return a list with one dictionary.  Why a list?
        # Some people feel a need in custom code to support a different adapter for
        # RelationList fields.  Such a field may point to three images.
        # In that case the adapter could return information for all three images,
        # so a list of three dictionaries.  The default case should use the same
        # structure.
        return [
            {
                "filename": image.filename,
                "content-type": image.contentType,
                "size": image.getSize(),
                "download": self._scale_view_from_url(url),
                "width": width,
                "height": height,
                "scales": scales,
            }
        ]

    def get_scales(self, field, width, height):
        """Get a dictionary of available scales for a particular image field,
        with the actual dimensions (aspect ratio of the original image).
        """
        scales = {}

        for name, actual_width, actual_height in _get_scale_infos():
            if actual_width > width:
                # The width of the scale is larger than the original width.
                # Scaling would simply return the original (or perhaps a copy
                # with the same size).  We do not need this scale.
                # If we *do* want this, we should call the scale method with
                # mode="cover", so it scales up.
                continue

            # Get the scale info without actually generating the scale,
            # nor any old-style HiDPI scales.
            scale = self.images_view.scale(
                field.__name__,
                width=actual_width,
                height=actual_height,
                pre=True,
                include_srcset=False,
            )
            if scale is None:
                # If we cannot get a scale, it is probably a corrupt image.
                continue

            url = scale.url
            actual_width = scale.width
            actual_height = scale.height

            scales[name] = {
                "download": self._scale_view_from_url(url),
                "width": actual_width,
                "height": actual_height,
            }

        return scales

    def get_original_image_url(self, fieldname, width, height):
        scale = self.images_view.scale(
            fieldname,
            width=width,
            height=height,
            pre=True,
            include_srcset=False,
        )
        # Corrupt images may not have a scale.
        return scale.url if scale else None

    def _scale_view_from_url(self, url):
        # The "download" information for scales is the path to
        # "@@images/foo-scale" only.
        # The full URL to the scale is rendered by the scaling adapter at
        # runtime to make sure they are correct in virtual hostings.
        return url.replace(self.context.absolute_url(), "").lstrip("/")



from plone.namedfile.file import NamedBlobFile as BlobFileValueType
from plone.namedfile.file import NamedBlobImage as BlobImageValueType
from plone.namedfile.file import NamedFile as FileValueType
from plone.namedfile.file import NamedImage as ImageValueType
from plone.namedfile.interfaces import INamedBlobFile
from plone.namedfile.interfaces import INamedBlobFileField
from plone.namedfile.interfaces import INamedBlobImage
from plone.namedfile.interfaces import INamedBlobImageField
from plone.namedfile.interfaces import INamedFile
from plone.namedfile.interfaces import INamedFileField
from plone.namedfile.interfaces import INamedImage
from plone.namedfile.interfaces import INamedImageField
from plone.namedfile.interfaces import IPluggableFileFieldValidation
from plone.namedfile.interfaces import IPluggableImageFieldValidation
from plone.namedfile.utils import get_contenttype
from zope.component import adapter
from zope.component import getAdapters
from zope.i18nmessageid import MessageFactory
from zope.interface import implementer
from zope.interface import Interface
from zope.schema import Object
from zope.schema import ValidationError

import mimetypes

_ = MessageFactory("plone")


class InvalidFile(ValidationError):
    """Exception for a invalid file."""

    __doc__ = _("Invalid file")


class InvalidImageFile(ValidationError):
    """Exception for a invalid image file."""

    __doc__ = _("Invalid image file")


class BinaryContenttypeValidator:
    def __init__(self, field, value):
        self.field = field
        self.value = value

    def __call__(self):
        if self.value is None:
            return

        if not self.field.accept:
            # No restrictions.
            return

        mimetype = get_contenttype(self.value)

        for accept in self.field.accept:
            if accept[0] == ".":
                # This is a file extension. Get a media type from it.
                accept = mimetypes.guess_type(f"dummy{accept}", strict=False)[0]
                if accept is None:
                    # This extension is unknown. Skip it.
                    continue

            try:
                accept_type, accept_subtype = accept.split("/")
                content_type, content_subtype = mimetype.split("/")
            except ValueError:
                # The accept type is invalid. Skip it.
                continue

            if accept_type == content_type and (
                accept_subtype == content_subtype or accept_subtype == "*"
            ):
                # This file is allowed, just don't raise a ValidationError.
                return

        # The file's content type is not allowed. Raise a ValidationError.
        raise self.exception(mimetype, self.field.__name__)


@implementer(IPluggableFileFieldValidation)
@adapter(INamedFileField, Interface)
class FileContenttypeValidator(BinaryContenttypeValidator):
    exception = InvalidFile


@implementer(IPluggableImageFieldValidation)
@adapter(INamedImageField, Interface)
class ImageContenttypeValidator(BinaryContenttypeValidator):
    exception = InvalidImageFile


class NamedField(Object):

    def __init__(self, **kw):
        if "accept" in kw:
            self.accept = kw.pop("accept")
        if "schema" in kw:
            self.schema = kw.pop("schema")
        super().__init__(schema=self.schema, **kw)

    def validate(self, value, interface):
        super().validate(value)
        for name, validator in getAdapters((self, value), interface):
            validator()


@implementer(INamedFileField)
class NamedFile(NamedField):
    """A NamedFile field"""

    _type = FileValueType
    schema = INamedFile
    accept = ()

    def validate(self, value):
        super().validate(value, IPluggableFileFieldValidation)


@implementer(INamedImageField)
class NamedImage(NamedField):
    """A NamedImage field"""

    _type = ImageValueType
    schema = INamedImage
    accept = ("image/*",)

    def validate(self, value):
        super().validate(value, IPluggableImageFieldValidation)


@implementer(INamedBlobFileField)
class NamedBlobFile(NamedField):
    """A NamedBlobFile field"""

    _type = BlobFileValueType
    schema = INamedBlobFile
    accept = ()

    def validate(self, value):
        super().validate(value, IPluggableFileFieldValidation)


@implementer(INamedBlobImageField)
class NamedBlobImage(NamedField):
    """A NamedBlobImage field"""

    _type = BlobImageValueType
    schema = INamedBlobImage
    accept = ("image/*",)

    def validate(self, value):
        super().validate(value, IPluggableImageFieldValidation)



# This file is partially borrowed from zope.app.file and licensed ZPL.

from DateTime import DateTime
from plone.namedfile.file import NamedImage
from plone.namedfile.interfaces import INamedImage
from plone.namedfile.tests import getFile
from plone.namedfile.tests import MockNamedImage
from plone.namedfile.utils import get_contenttype
from zope.interface.verify import verifyClass

import time
import unittest


zptlogo = (
    b"GIF89a\x10\x00\x10\x00\xd5\x00\x00\xff\xff\xff\xff\xff\xfe\xfc\xfd\xfd"
    b"\xfa\xfb\xfc\xf7\xf9\xfa\xf5\xf8\xf9\xf3\xf6\xf8\xf2\xf5\xf7\xf0\xf4\xf6"
    b"\xeb\xf1\xf3\xe5\xed\xef\xde\xe8\xeb\xdc\xe6\xea\xd9\xe4\xe8\xd7\xe2\xe6"
    b"\xd2\xdf\xe3\xd0\xdd\xe3\xcd\xdc\xe1\xcb\xda\xdf\xc9\xd9\xdf\xc8\xd8\xdd"
    b"\xc6\xd7\xdc\xc4\xd6\xdc\xc3\xd4\xda\xc2\xd3\xd9\xc1\xd3\xd9\xc0\xd2\xd9"
    b"\xbd\xd1\xd8\xbd\xd0\xd7\xbc\xcf\xd7\xbb\xcf\xd6\xbb\xce\xd5\xb9\xcd\xd4"
    b"\xb6\xcc\xd4\xb6\xcb\xd3\xb5\xcb\xd2\xb4\xca\xd1\xb2\xc8\xd0\xb1\xc7\xd0"
    b"\xb0\xc7\xcf\xaf\xc6\xce\xae\xc4\xce\xad\xc4\xcd\xab\xc3\xcc\xa9\xc2\xcb"
    b"\xa8\xc1\xca\xa6\xc0\xc9\xa4\xbe\xc8\xa2\xbd\xc7\xa0\xbb\xc5\x9e\xba\xc4"
    b"\x9b\xbf\xcc\x98\xb6\xc1\x8d\xae\xbaFgs\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00,\x00\x00\x00\x00\x10\x00\x10\x00\x00\x06z@\x80pH,\x12k\xc8$\xd2f\x04"
    b"\xd4\x84\x01\x01\xe1\xf0d\x16\x9f\x80A\x01\x91\xc0ZmL\xb0\xcd\x00V\xd4"
    b"\xc4a\x87z\xed\xb0-\x1a\xb3\xb8\x95\xbdf8\x1e\x11\xca,MoC$\x15\x18{"
    b"\x006}m\x13\x16\x1a\x1f\x83\x85}6\x17\x1b $\x83\x00\x86\x19\x1d!%)\x8c"
    b"\x866#'+.\x8ca`\x1c`(,/1\x94B5\x19\x1e\"&*-024\xacNq\xba\xbb\xb8h\xbeb"
    b"\x00A\x00;"
)


class TestImage(unittest.TestCase):
    def _makeImage(self, *args, **kw):
        return NamedImage(*args, **kw)

    def testEmpty(self):
        file_img = self._makeImage()
        self.assertEqual(file_img.contentType, "")
        self.assertEqual(bytes(file_img.data), b"")
        self.assertIsNotNone(file_img.modified)

    def testConstructor(self):
        file_img = self._makeImage(b"Data")
        self.assertEqual(file_img.contentType, "")
        self.assertEqual(bytes(file_img.data), b"Data")
        self.assertIsNotNone(file_img.modified)

    def testMutators(self):
        image = self._makeImage()
        image.contentType = "image/jpeg"
        self.assertEqual(image.contentType, "image/jpeg")

        image._setData(zptlogo)
        self.assertEqual(image.data, zptlogo)
        self.assertEqual(image.contentType, "image/gif")
        self.assertEqual(image.getImageSize(), (16, 16))

    def testModifiedTimeStamp(self):
        image = self._makeImage()
        old_timestamp = image.modified
        time.sleep(1/1000)  # make sure at least 1ms passes
        image._setData(zptlogo)
        self.assertNotEqual(image.modified, old_timestamp)

    def testFallBackToDatabaseModifiedTimeStamp(self):
        dt = DateTime()
        image = MockNamedImage()
        image._p_mtime = dt.millis()
        image._modified = (dt + 1).millis()

        delattr(image, "_modified")
        marker = object()
        self.assertEqual(marker, getattr(image, "_modified", marker))
        self.assertEqual(dt.millis(), image._p_mtime)

    def testInterface(self):
        self.assertTrue(INamedImage.implementedBy(NamedImage))
        self.assertTrue(verifyClass(INamedImage, NamedImage))

    def test_extract_media_type(self):
        from plone.namedfile.utils import extract_media_type as extract

        self.assertIsNone(extract(None))
        self.assertEqual(extract("text/plain"), "text/plain")
        self.assertEqual(extract("TEXT/PLAIN"), "text/plain")
        self.assertEqual(extract("text / plain"), "text/plain")
        self.assertEqual(extract(" text/plain ; charset=utf-8"), "text/plain")



from plone.namedfile.file import NamedImage
from plone.namedfile.tests import getFile
from plone.namedfile.utils import get_contenttype

import unittest


class TestUtils(unittest.TestCase):

    def test_get_contenttype(self):
        self.assertEqual(
            get_contenttype(
                NamedImage(
                    getFile("image.gif"),
                    contentType="image/gif",
                )
            ),
            "image/gif",
        )
        self.assertEqual(
            get_contenttype(
                NamedImage(
                    getFile("image.gif"),
                    filename="image.gif",
                )
            ),
            "image/gif",
        )
        self.assertEqual(
            get_contenttype(
                NamedImage(
                    getFile("image.tif"),
                    filename="image.tif",
                )
            ),
            "image/tiff",
        )
        self.assertEqual(
            get_contenttype(
                NamedImage(
                    getFile("notimage.doc"),
                    filename="notimage.doc",
                )
            ),
            "application/msword",
        )

        # Filename only detection of a non-IANA registered type.
        self.assertEqual(
            get_contenttype(filename="image.webp"),
            "image/webp",
        )

        # Filename only detection of a non-IANA registered type.
        self.assertEqual(
            get_contenttype(filename="song.midi"),
            "audio/midi",
        )

        # Detection of a surely not registered type.
        self.assertEqual(
            get_contenttype(filename="nothing.plonenamedfile"),
            "application/octet-stream",
        )



from DateTime import DateTime
from plone.namedfile.file import NamedBlobImage
from plone.namedfile.file import NamedImage


import os


def getFile(filename, length=None):
    """return contents of the file with the given name"""
    filename = os.path.join(os.path.dirname(__file__), filename)
    with open(filename, "rb") as data_file:
        return data_file.read(length)


class MockNamedImage(NamedImage):
    _p_mtime = DateTime().millis()


class MockNamedBlobImage(NamedBlobImage):
    _p_mtime = DateTime().millis()



# This file was borrowed from z3c.blobfile and is licensed under the terms of
# the ZPL.

##############################################################################
#
# Copyright (c) 2008 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

from OFS.Image import Pdata
from plone.namedfile.file import FileChunk
from plone.namedfile.file import NamedBlobImage
from plone.namedfile.testing import PLONE_NAMEDFILE_FUNCTIONAL_TESTING
from plone.namedfile.tests import getFile

import os
import tempfile
import unittest


class TestStorable(unittest.TestCase):

    layer = PLONE_NAMEDFILE_FUNCTIONAL_TESTING

    def test_pdata_storable(self):
        pdata = Pdata(getFile("image.gif"))
        fi = NamedBlobImage(pdata, filename="image.gif")
        self.assertEqual(303, fi.getSize())

    def test_str_storable(self):
        fi = NamedBlobImage(getFile("image.gif"), filename="image.gif")
        self.assertEqual(303, fi.getSize())

    def test_filechunk_storable(self):
        fi = NamedBlobImage(FileChunk(getFile("image.gif")), filename="image.gif")
        self.assertEqual(303, fi.getSize())

    def test_opened_file_storable(self):
        data = getFile("image.gif")
        f = tempfile.NamedTemporaryFile(delete=False)
        try:
            path = f.name
            f.write(data)
            f.close()
            with open(path, "rb") as f:
                fi = NamedBlobImage(f, filename="image.gif")
        finally:
            if os.path.exists(path):
                os.remove(path)
        self.assertEqual(303, fi.getSize())



# This file was borrowed from z3c.blobfile and is licensed under the terms of
# the ZPL.
from OFS.Image import Pdata
from plone.namedfile.file import FileChunk
from plone.namedfile.interfaces import IStorage
from plone.namedfile.interfaces import NotStorable
from zope.interface import implementer
from zope.publisher.browser import FileUpload

import io


MAXCHUNKSIZE = 1 << 16


@implementer(IStorage)
class BytesStorable:
    def store(self, data, blob):
        if not isinstance(data, bytes):
            raise NotStorable("Could not store data (not of bytes type).")

        with blob.open("w") as fp:
            fp.write(data)


@implementer(IStorage)
class TextStorable(BytesStorable):
    def store(self, data, blob):
        if not isinstance(data, str):
            raise NotStorable('Could not store data (not of "unicode" type).')

        data = data.encode("UTF-8")
        BytesStorable.store(self, data, blob)


@implementer(IStorage)
class UnicodeStorable(TextStorable):
    pass


@implementer(IStorage)
class StringStorable(BytesStorable):
    pass


@implementer(IStorage)
class FileChunkStorable:
    def store(self, data, blob):
        if not isinstance(data, FileChunk):
            raise NotStorable(
                'Could not store data (not a of "FileChunk" type).'
            )  # noqa

        with blob.open("w") as fp:
            chunk = data
            while chunk:
                fp.write(chunk._data)
                chunk = chunk.next


@implementer(IStorage)
class FileDescriptorStorable:
    def store(self, data, blob):
        if not isinstance(data, io.IOBase):
            raise NotStorable('Could not store data: not of io.IOBase ("file").')

        filename = getattr(data, "name", None)
        if filename is not None:
            blob.consumeFile(filename)
            return


class BufferedReaderStorable:
    def store(self, data, blob):
        raw = data.raw
        if not isinstance(raw, io.FileIO):
            raise NotStorable('Could not store data (not of type "io.FileIO")')

        filename = getattr(data.raw, "name", None)
        if filename is not None:
            blob.consumeFile(filename)
            return


@implementer(IStorage)
class FileUploadStorable:
    def store(self, data, blob):
        if not isinstance(data, FileUpload):
            raise NotStorable('Could not store data (not of "FileUpload").')

        data.seek(0)

        with blob.open("w") as fp:
            block = data.read(MAXCHUNKSIZE)
            while block:
                fp.write(block)
                block = data.read(MAXCHUNKSIZE)


@implementer(IStorage)
class PDataStorable:
    def store(self, pdata, blob):
        if not isinstance(pdata, Pdata):
            raise NotStorable('Could not store data (not of "Pdata").')
        fp = blob.open("w")
        fp.write(bytes(pdata))
        fp.close()


@implementer(IStorage)
class Zope2FileUploadStorable:
    def store(self, data, blob):
        data.seek(0)
        with blob.open("w") as fp:
            block = data.read(MAXCHUNKSIZE)
            while block:
                fp.write(block)
                block = data.read(MAXCHUNKSIZE)



# The implementations in this file are largely borrowed
# from zope.app.file and z3c.blobfile
# and are licensed under the ZPL.
from DateTime import DateTime
from logging import getLogger
from persistent import Persistent
from plone.namedfile.interfaces import INamedBlobFile
from plone.namedfile.interfaces import INamedBlobImage
from plone.namedfile.interfaces import INamedFile
from plone.namedfile.interfaces import INamedImage
from plone.namedfile.interfaces import IStorage
from plone.namedfile.utils import get_contenttype
from plone.namedfile.utils import get_exif
from plone.namedfile.utils import getImageInfo
from plone.namedfile.utils import rotate_image
from ZODB.blob import Blob
from zope.component import getUtility
from zope.interface import implementer
from zope.schema.fieldproperty import FieldProperty
from ZPublisher import HTTPRangeSupport

import piexif
import transaction


log = getLogger(__name__)


MAXCHUNKSIZE = 1 << 16
IMAGE_INFO_BYTES = 1024
MAX_INFO_BYTES = 1 << 18


class FileChunk(Persistent):
    """Wrapper for possibly large data"""

    next = None

    def __init__(self, data):
        self._data = data

    def __getslice__(self, i, j):
        return self._data[i:j]

    def __len__(self):
        data = bytes(self)
        return len(data)

    def _get_contents(self):
        next = self.next
        if next is None:
            return self._data

        result = [self._data]
        while next is not None:
            self = next
            result.append(self._data)
            next = self.next

        return b"".join(result)

    __bytes__ = _get_contents


FILECHUNK_CLASSES = [FileChunk]
try:
    from zope.app.file.file import FileChunk as zafFileChunk

    FILECHUNK_CLASSES.append(zafFileChunk)
except ImportError:
    pass


class ModifiedPropertyMixin:
    @property
    def modified(self):
        if hasattr(self, "_modified"):
            return self._modified / 1000
        # Fall back to modification time in database.
        return self._p_mtime


@implementer(INamedFile)
class NamedFile(Persistent, ModifiedPropertyMixin):
    """A non-BLOB file that stores a filename

    Let's test the constructor:

    >>> file = NamedFile()
    >>> file.contentType
    ''
    >>> file.data
    ''

    >>> file = NamedFile('Foobar')
    >>> file.contentType
    ''
    >>> file.data
    'Foobar'

    >>> file = NamedFile('Foobar', 'text/plain')
    >>> file.contentType
    'text/plain'
    >>> file.data
    'Foobar'

    >>> file = NamedFile(data='Foobar', contentType='text/plain')
    >>> file.contentType
    'text/plain'
    >>> file.data
    'Foobar'


    Let's test the mutators:

    >>> file = NamedFile()
    >>> file.contentType = 'text/plain'
    >>> file.contentType
    'text/plain'

    >>> file.data = 'Foobar'
    >>> file.data
    'Foobar'

    >>> file.data = None
    Traceback (most recent call last):
    ...
    TypeError: Cannot set None data on a file.


    Let's test large data input:

    >>> file = NamedFile()

    Insert as string:

    >>> file.data = 'Foobar'*60000
    >>> file.getSize()
    360000
    >>> file.data == 'Foobar'*60000
    True

    Insert data as FileChunk:

    >>> fc = FileChunk('Foobar'*4000)
    >>> file.data = fc
    >>> file.getSize()
    24000
    >>> file.data == 'Foobar'*4000
    True

    Insert data from file object:

    >>> import cStringIO
    >>> sio = cStringIO.StringIO()
    >>> sio.write('Foobar'*100000)
    >>> sio.seek(0)
    >>> file.data = sio
    >>> file.getSize()
    600000
    >>> file.data == 'Foobar'*100000
    True


    Last, but not least, verify the interface:

    >>> from zope.interface.verify import verifyClass
    >>> INamedFile.implementedBy(NamedFile)
    True
    >>> verifyClass(INamedFile, NamedFile)
    True
    """

    filename = FieldProperty(INamedFile["filename"])

    def __init__(self, data=b"", contentType="", filename=None):
        if filename is not None and contentType in ("", "application/octet-stream"):
            contentType = get_contenttype(filename=filename)
        self.data = data
        self.contentType = contentType
        self.filename = filename
        self._modified = DateTime().millis()

    def _getData(self):
        if isinstance(self._data, tuple(FILECHUNK_CLASSES)):
            return bytes(self._data)
        else:
            return self._data

    def _setData(self, data):
        self._modified = DateTime().millis()

        # Handle case when data is a string
        if isinstance(data, str):
            data = data.encode("UTF-8")

        if isinstance(data, bytes):
            self._data, self._size = FileChunk(data), len(data)
            return

        # Handle case when data is None
        if data is None:
            raise TypeError("Cannot set None data on a file.")

        # Handle case when data is already a FileChunk
        if isinstance(data, tuple(FILECHUNK_CLASSES)):
            size = len(data)
            self._data, self._size = data, size
            return

        # Handle case when data is a file object
        seek = data.seek
        read = data.read

        seek(0, 2)
        size = end = data.tell()

        if size <= 2 * MAXCHUNKSIZE:
            seek(0)
            if size < MAXCHUNKSIZE:
                self._data, self._size = read(size), size
                return
            self._data, self._size = FileChunk(read(size)), size
            return

        # Make sure we have an _p_jar, even if we are a new object, by
        # doing a sub-transaction commit.
        transaction.savepoint(optimistic=True)

        jar = self._p_jar

        if jar is None:
            # Ugh
            seek(0)
            self._data, self._size = FileChunk(read(size)), size
            return

        # Now we're going to build a linked list from back
        # to front to minimize the number of database updates
        # and to allow us to get things out of memory as soon as
        # possible.
        next = None
        while end > 0:
            pos = end - MAXCHUNKSIZE
            if pos < MAXCHUNKSIZE:
                pos = 0  # we always want at least MAXCHUNKSIZE bytes
            seek(pos)
            data = FileChunk(read(end - pos))

            # Woooop Woooop Woooop! This is a trick.
            # We stuff the data directly into our jar to reduce the
            # number of updates necessary.
            jar.add(data)

            # This is needed and has side benefit of getting
            # the thing registered:
            data.next = next

            # Now make it get saved in a sub-transaction!
            transaction.savepoint(optimistic=True)

            # Now make it a ghost to free the memory.  We
            # don't need it anymore!
            data._p_changed = None

            next = data
            end = pos

        self._data, self._size = next, size
        return

    def getSize(self):
        """See `IFile`"""
        return self._size

    # See IFile.
    data = property(_getData, _setData)


@implementer(INamedImage)
class NamedImage(NamedFile):
    """An non-BLOB image with a filename"""

    filename = FieldProperty(INamedFile["filename"])

    def __init__(self, data=b"", contentType="", filename=None):
        self.contentType, self._width, self._height = getImageInfo(data)
        self.filename = filename
        self._setData(data)

        # Allow override of the image sniffer
        if contentType:
            self.contentType = contentType

        exif_data = get_exif(data)
        if exif_data is not None:
            log.debug(
                "Image contains Exif Informations. "
                "Test for Image Orientation and Rotate if necessary."
                "Exif Data: %s",
                exif_data,
            )
            orientation = exif_data["0th"].get(piexif.ImageIFD.Orientation, 1)
            if 1 < orientation <= 8:
                self.data, self._width, self._height, self.exif = rotate_image(
                    self.data
                )
            self.exif_data = exif_data

    def _setData(self, data):
        super()._setData(data)

        contentType, self._width, self._height = getImageInfo(self._data)
        if contentType:
            self.contentType = contentType

    def getImageSize(self):
        """See interface `IImage`"""
        return (self._width, self._height)

    data = property(NamedFile._getData, _setData)


@implementer(INamedBlobFile, HTTPRangeSupport.HTTPRangeInterface)
class NamedBlobFile(Persistent, ModifiedPropertyMixin):
    """A file stored in a ZODB BLOB, with a filename"""

    filename = FieldProperty(INamedFile["filename"])

    def __init__(self, data=b"", contentType="", filename=None):
        if filename is not None and contentType in ("", "application/octet-stream"):
            contentType = get_contenttype(filename=filename)
        self.contentType = contentType
        self._blob = Blob()
        f = self._blob.open("w")
        f.write(b"")
        f.close()
        self._setData(data)
        self.filename = filename
        self._modified = DateTime().millis()

    def open(self, mode="r"):
        if mode != "r" and "size" in self.__dict__:
            del self.__dict__["size"]
        return self._blob.open(mode)

    def openDetached(self):
        return open(self._blob.committed(), "rb")

    def _setData(self, data):
        if "size" in self.__dict__:
            del self.__dict__["size"]
        # Search for a storable that is able to store the data
        dottedName = ".".join((data.__class__.__module__, data.__class__.__name__))
        log.debug("Storage selected for data: %s", dottedName)
        storable = getUtility(IStorage, name=dottedName)
        storable.store(data, self._blob)
        self._modified = DateTime().millis()

    def _getData(self):
        fp = self._blob.open("r")
        data = fp.read()
        fp.close()
        return data

    _data = property(_getData, _setData)
    data = property(_getData, _setData)

    @property
    def size(self):
        if "size" in self.__dict__:
            return self.__dict__["size"]
        with self._blob.open() as reader:
            reader.seek(0, 2)
            size = int(reader.tell())
        self.__dict__["size"] = size
        return size

    def getSize(self):
        return self.size


@implementer(INamedBlobImage)
class NamedBlobImage(NamedBlobFile):
    """An image stored in a ZODB BLOB with a filename"""

    def __init__(self, data=b"", contentType="", filename=None):
        super().__init__(data, contentType=contentType, filename=filename)

        # Allow override of the image sniffer
        if contentType:
            self.contentType = contentType
        exif_data = get_exif(self.data)
        if exif_data is not None:
            log.debug(
                "Image contains Exif Informations. "
                "Test for Image Orientation and Rotate if necessary."
                "Exif Data: %s",
                exif_data,
            )
            orientation = exif_data["0th"].get(piexif.ImageIFD.Orientation, 1)
            if 1 < orientation <= 8:
                try:
                    self.data, self._width, self._height, self.exif = rotate_image(
                        self.data
                    )
                except KeyboardInterrupt:
                    raise
                except Exception:
                    log.warning(
                        "Error rotating image %s based on exif data.",
                        filename,
                        exc_info=1,
                    )
            else:
                self.exif = exif_data

    def _setData(self, data):
        super()._setData(data)
        firstbytes = self.getFirstBytes()
        res = getImageInfo(firstbytes)
        if res == ("image/jpeg", -1, -1) or res == ("image/tiff", -1, -1) or \
                res == ("image/svg+xml", -1, -1):
            # header was longer than firstbytes
            start = len(firstbytes)
            length = max(0, MAX_INFO_BYTES - start)
            firstbytes += self.getFirstBytes(start, length)
            res = getImageInfo(firstbytes)
        contentType, self._width, self._height = res
        if contentType:
            self.contentType = contentType

    data = property(NamedBlobFile._getData, _setData)

    def getFirstBytes(self, start=0, length=IMAGE_INFO_BYTES):
        """Returns the first bytes of the file.

        Returns an amount which is sufficient to determine the image type.
        """
        with self.open("r") as fp:
            fp.seek(start)
            firstbytes = fp.read(length)
        return firstbytes

    def getImageSize(self):
        """See interface `IImage`"""
        if (self._width, self._height) != (-1, -1):
            return (self._width, self._height)

        contentType, self._width, self._height = getImageInfo(self.data)
        return (self._width, self._height)



from bs4 import BeautifulSoup
from plone.app.uuid.utils import uuidToObject
from plone.namedfile.interfaces import IAvailableSizes
from plone.registry.interfaces import IRegistry
from zope.component import getUtility
from zope.component import queryUtility

import logging
import re


try:
    from plone.base.interfaces import IImagingSchema
except ImportError:
    from Products.CMFPlone.interfaces.controlpanel import IImagingSchema


logger = logging.getLogger(__name__)
appendix_re = re.compile("^(.*)([?#].*)$")
resolveuid_re = re.compile("^[./]*resolve[Uu]id/([^/]*)/?(.*)$")


def get_allowed_scales():
    sizes_util = queryUtility(IAvailableSizes)
    if sizes_util is None:
        return {}
    sizes = sizes_util()
    if sizes is None:
        return {}
    return sizes


def get_picture_variants():
    registry = getUtility(IRegistry)
    settings = registry.forInterface(IImagingSchema, prefix="plone", check=False)
    return getattr(settings, "picture_variants", {})


class Img2PictureTag:
    def get_scale_name(self, scale_line):
        parts = scale_line.split(" ")
        return parts and parts[0] or ""

    def get_scale_width(self, scale):
        """get width from allowed_scales line
        large 800:65536
        """
        allowed_scales = get_allowed_scales()
        scale_info = allowed_scales.get(scale)
        if not scale_info:
            return
        return scale_info[0]

    def create_picture_tag(
        self, sourceset, attributes, uid=None, fieldname=None, resolve_urls=False
    ):
        """Converts the img tag to a picture tag with picture_variant definition"""
        width = None
        height = None
        src = attributes.get("src")
        if not uid and not src:
            raise TypeError("Either uid or attributes['src'] need to be given.")
        soup = BeautifulSoup("", "html.parser")
        allowed_scales = get_allowed_scales()
        if uid:
            obj = uuidToObject(uid, unrestricted=True)
        else:
            obj = self.resolve_uid_url(src)
        picture_tag = soup.new_tag("picture")
        css_classes = attributes.get("class", [])
        if "captioned" in css_classes:
            picture_tag["class"] = "captioned"
        for i, source in enumerate(sourceset):
            target_scale = source["scale"]
            media = source.get("media")

            additional_scales = source.get("additionalScales", None)
            if additional_scales is None:
                additional_scales = [
                    self.get_scale_name(s) for s in allowed_scales if s != target_scale
                ]
            source_scales = [target_scale] + additional_scales
            source_srcset = []
            for scale in source_scales:
                scale_width = self.get_scale_width(scale)
                if not scale_width:
                    logger.warning("No width found for scale %s.", scale)
                    continue
                if resolve_urls and obj:
                    scale_view = obj.unrestrictedTraverse("@@images", None)
                    scale_obj = scale_view.scale(fieldname, scale, pre=True)
                    scale_url = scale_obj.url
                else:
                    # obj = self.resolve_uid_url(src)
                    # scale_view = obj.unrestrictedTraverse("@@images", None)
                    # scale_obj = scale_view.scale(fieldname, scale, pre=True)
                    # scale_url = scale_obj.url
                    scale_url = self.update_src_scale(src=src, scale=scale)
                source_srcset.append(f"{scale_url} {scale_width}w")
            source_tag = soup.new_tag("source", srcset=",\n".join(source_srcset))
            if media:
                source_tag["media"] = media
            picture_tag.append(source_tag)
            if i == len(sourceset) - 1:
                if resolve_urls and obj:
                    scale_view = obj.unrestrictedTraverse("@@images", None)
                    scale_obj = scale_view.scale(fieldname, target_scale, pre=True)
                    scale_url = scale_obj.url
                    width = scale_obj.width
                    height = scale_obj.height
                else:
                    # obj = self.resolve_uid_url(src)
                    # scale_view = obj.unrestrictedTraverse("@@images", None)
                    # scale_obj = scale_view.scale(fieldname, target_scale, pre=True)
                    # scale_url = scale_obj.url
                    # width = scale_obj.width
                    # height = scale_obj.height
                    scale_url = self.update_src_scale(src=src, scale=target_scale)
                img_tag = soup.new_tag("img", src=scale_url)
                for k, attr in attributes.items():
                    if k in ["src", "srcset"]:
                        continue
                    img_tag.attrs[k] = attr
                img_tag["loading"] = "lazy"
                if width:
                    img_tag["width"] = width
                if height:
                    img_tag["height"] = height
                picture_tag.append(img_tag)
        return picture_tag

    def resolve_uid_url(self, href):
        obj = None
        subpath = href
        match = resolveuid_re.match(subpath)
        if match is not None:
            uid, _subpath = match.groups()
            obj = uuidToObject(uid, unrestricted=True)
        return obj

    def update_src_scale(self, src, scale):
        parts = src.split("/")
        if "." in parts[-1]:
            field_name = parts[-1].split("-")[0]
            src_scale = "/".join(parts[:-1]) + f"/{field_name}/{scale}"
            src_scale
        else:
            # Usually the url has '@@images/fieldname/other_scale',
            # and then we replace the other scale.
            # But the url may use the original image, e.g. @@images/image.
            # Then we want to keep the fieldname and return '.../image/scale'.
            try:
                full = len(parts) - parts.index("@@images") == 2
            except ValueError:
                full = False
            if not full:
                parts = parts[:-1]
            src_scale = "/".join(parts) + f"/{scale}"
        return src_scale



from io import BytesIO
from ZPublisher.HTTPRequest import FileUpload

import os


GIF_FILE = os.path.join(os.path.dirname(__file__), "tool.gif")
with open(GIF_FILE, "rb") as f:
    GIF = f.read()
# jpeg file of 900x900 pixels
JPEG_FILE = os.path.join(os.path.dirname(__file__), "900.jpg")
with open(JPEG_FILE, "rb") as f:
    JPEG = f.read()


class File(FileUpload):
    """Dummy upload object
    Used to fake uploaded files.
    """

    __allow_access_to_unprotected_subobjects__ = 1
    filename = "dummy.txt"
    data = b"file data"
    headers = {}

    def __init__(self, filename=None, data=None, headers=None):
        if filename is not None:
            self.filename = filename
        if data is not None:
            self.data = data
        if headers is not None:
            self.headers = headers
        self.file = BytesIO(self.data)

    def seek(self, *args):
        pass

    def tell(self, *args):
        return 1

    def read(self, *args):
        return self.data


class Image(File):
    """Dummy image upload object
    Contains valid image data by default.
    """

    filename = "dummy.gif"
    data = GIF


class JpegImage(File):
    """Dummy jpeg image upload object

    900 by 900 pixels.
    """

    filename = "900.jpeg"
    data = JPEG



from DateTime import DateTime
from doctest import _ellipsis_match
from io import BytesIO
from OFS.SimpleItem import SimpleItem
from plone.namedfile.field import NamedImage as NamedImageField
from plone.namedfile.file import NamedImage
from plone.namedfile.interfaces import IAvailableSizes
from plone.namedfile.interfaces import IImageScaleTraversable
from plone.namedfile.scaling import ImageScale
from plone.namedfile.scaling import ImageScaling
from plone.namedfile.testing import PLONE_NAMEDFILE_FUNCTIONAL_TESTING
from plone.namedfile.testing import PLONE_NAMEDFILE_INTEGRATION_TESTING
from plone.namedfile.tests import getFile
from plone.namedfile.tests import MockNamedImage
from plone.rfc822.interfaces import IPrimaryFieldInfo
from plone.scale.interfaces import IScaledImageQuality
from plone.scale.storage import IImageScaleStorage
from unittest.mock import patch
from zExceptions import Unauthorized
from zope.annotation import IAttributeAnnotatable
from zope.component import adapter
from zope.component import getGlobalSiteManager
from zope.component import getSiteManager
from zope.interface import implementer
from zope.interface import Interface
from zope.publisher.interfaces import NotFound

import PIL
import plone.namedfile.picture
import plone.namedfile.scaling
import re
import time
import unittest
import warnings


# Unique scale name used to be a uuid.uui4(),
# which is a combination of hexadecimal digits with dashes, total 36.
# Now it is 'imagescalename-width-hash', where hash is 32.
PAT_UID_SCALE = r"[0-9a-z]*-[0-9]*-[0-9a-f]{32}"


def wait_to_ensure_modified():
    # modified is measured in milliseconds
    # wait 5ms to ensure modified will have changed
    time.sleep(0.005)


class IHasImage(IImageScaleTraversable):
    image = NamedImageField()


def assertImage(testcase, data, format_, size):
    image = PIL.Image.open(BytesIO(data))
    testcase.assertEqual(image.format, format_)
    testcase.assertEqual(image.size, size)


def patch_Img2PictureTag_picture_variants():

    return {
        "large": {
            "title": "Large",
            "sourceset": [
                {
                    "scale": "larger",
                    "additionalScales": [
                        "preview",
                        "teaser",
                        "large",
                        "great",
                        "huge",
                    ],
                }
            ],
        },
        "medium": {
            "title": "Medium",
            "sourceset": [
                {
                    "scale": "teaser",
                    "additionalScales": ["preview", "large", "larger", "great"],
                }
            ],
        },
        "small": {
            "title": "Small",
            "sourceset": [
                {
                    "scale": "preview",
                    "additionalScales": ["preview", "large", "larger"],
                }
            ],
        },
    }


def patch_Img2PictureTag_empty_picture_variants():
    # You would have this in Plone 5.2, or if someone empties the registry setting.
    return {}


def patch_Img2PictureTag_allowed_scales():

    return {
        "huge": (1600, 65536),
        "great": (1200, 65536),
        "larger": (1000, 65536),
        "large": (800, 65536),
        "teaser": (600, 65536),
        "preview": (400, 65536),
        "mini": (200, 65536),
        "thumb": (128, 128),
        "tile": (64, 64),
        "icon": (32, 32),
        "listing": (16, 16),
    }


@implementer(IAttributeAnnotatable, IHasImage)
class DummyContent(SimpleItem):
    image = None
    modified = DateTime
    id = __name__ = "item"
    title = "foo"

    def Title(self):
        return self.title

    def UID(self):
        return "dummy_uuid"


@implementer(IPrimaryFieldInfo)
@adapter(DummyContent)
class PrimaryFieldInfo:
    def __init__(self, context):
        self.context = context
        self.fieldname = "image"
        self.field = self.context.image

    @property
    def value(self):
        return self.field


@implementer(IScaledImageQuality)
class DummyQualitySupplier:
    """fake utility for image quality setting from imaging control panel."""

    def getQuality(self):
        return 1  # as bad as it gets


class FakeImage:
    def __init__(self, value, format, key=None, uid="image"):
        self.value = value
        self.format = format
        self._width = len(value)
        self._height = len(format)
        self.contentType = f"image/{format}"
        # variables for scales:
        self._scales = {}
        self.key = key
        self.uid = uid

    @property
    def data(self):
        return self

    @property
    def info(self):
        return dict(
            data=self.data,
            width=self._width,
            height=self._height,
            mimetype=f"image/{self.format.lower()}",
            key=self.key,
            uid=self.uid,
        )

    def absolute_url(self):
        return "http://fake.image"

    def Title(self):
        # used for tag
        return "Image Title"


@implementer(IImageScaleStorage)
class FakeImageScaleStorage:
    """Storage class for FakeImages."""

    def __init__(self, context, modified=None):
        """Adapt the given context item and optionally provide a callable
        to return a representation of the last modification date, which
        can be used to invalidate stored scale data on update."""
        self.context = context
        self.modified = modified
        self.storage = context._scales

    def pre_scale(self, factory=None, **parameters):
        """Find image scale data for the given parameters or pre-create it.

        In our version, we only support height and width.
        """
        stripped_parameters = {
            "target_height": parameters.get("height"),
            "target_width": parameters.get("width"),
        }
        key = self.hash(**stripped_parameters)
        info = self.get_info_by_hash(key)
        if info is not None:
            # Note: we could do something with self.modified here,
            # but we choose to ignore it.
            return info
        return self.create_scale(no_scale=True, **stripped_parameters)

    def scale(self, factory=None, **parameters):
        """Find image scale data for the given parameters or create it.

        In our version, we only support height and width.
        """
        stripped_parameters = {
            "target_height": parameters.get("height"),
            "target_width": parameters.get("width"),
        }
        key = self.hash(**stripped_parameters)
        info = self.get_info_by_hash(key)
        if info is not None:
            # Note: we could do something with self.modified here,
            # but we choose to ignore it.
            return info
        return self.create_scale(**stripped_parameters)

    def create_scale(self, target_height=None, target_width=None, no_scale=False):
        if target_height is None and target_width is None:
            # Return the original.
            return self.context.info

        # We have a funny way of scaling.
        value = self.context.value[:target_height]
        format = self.context.format[:target_width]

        # Get uid and key.
        # Our implementation never throws away scales, so we can use uid-0, uid-1, etc.
        # Note: in ImageScaling.publishTraverse a dash in the url means it is a scale uid.
        uid = f"uid-{len(self.storage)}"
        key = self.hash(target_height=target_height, target_width=target_width)

        if no_scale:
            info = dict(
                data=None,
                width=self._width,
                height=self._height,
                mimetype=f"image/{self.format.lower()}",
                key=self.key,
                uid=self.uid,
            )
        else:
            # Create a new fake image for this scale.
            scale = FakeImage(value, format, key=key, uid=uid)
            info = scale.info

        # Store the real scale or placeholder scale and return the info.
        self.storage[uid] = info
        return scale.info

    def __getitem__(self, uid):
        """Find image scale data based on its uid."""
        return self.storage[uid]

    def get(self, uid, default=None):
        return self.storage.get(uid, default)

    def get_or_generate(self, uid, default=None):
        info = self.storage.get(uid, default)
        if info is None:
            return
        if info.get("data"):
            return info
        # We have a placeholder. Get real data.
        stripped_parameters = {
            "target_height": info.get("height"),
            "target_width": info.get("width"),
        }
        return self.create_scale(**stripped_parameters)

    def hash(self, **parameters):
        return tuple(parameters.values())

    def get_info_by_hash(self, hash):
        for value in self.storage.values():
            if value["key"] == hash:
                return value


class TitleImageScale(ImageScale):
    """ImageScale class with its own title property.
    """

    title = "title from class"


# @patch.multiple(
#     "plone.namedfile.scaling.Img2PictureTag",
#     allowed_scales=patch_Img2PictureTag_allowed_scales,
#     picture_variants=patch_Img2PictureTag_picture_variants,
#     spec=True,
# )
class ImageScalingTests(unittest.TestCase):

    layer = PLONE_NAMEDFILE_INTEGRATION_TESTING

    def setUp(self):
        sm = getSiteManager()
        sm.registerAdapter(PrimaryFieldInfo)

        data = getFile("image.png")
        item = DummyContent()
        item.image = MockNamedImage(data, "image/png", "image.png")
        self.layer["app"]._setOb("item", item)
        self.item = self.layer["app"].item
        self._orig_sizes = ImageScaling._sizes
        self.scaling = ImageScaling(self.item, None)

    def tearDown(self):
        ImageScaling._sizes = self._orig_sizes
        sm = getSiteManager()
        sm.unregisterAdapter(PrimaryFieldInfo)

    def testCreateScale(self):
        foo = self.scaling.scale("image", width=100, height=80)
        self.assertTrue(foo.uid)
        self.assertEqual(foo.mimetype, "image/png")
        self.assertIsInstance(foo.mimetype, str)
        self.assertEqual(foo.data.contentType, "image/png")
        self.assertIsInstance(foo.data.contentType, str)
        self.assertEqual(foo.width, 80)
        self.assertEqual(foo.height, 80)
        assertImage(self, foo.data.data, "PNG", (80, 80))

    def testCreateExactScale(self):
        foo = self.scaling.scale("image", width=100, height=80)
        self.assertIsNot(foo.data, self.item.image)

        # test that exact scale without parameters returns original
        foo = self.scaling.scale(
            "image", width=self.item.image._width, height=self.item.image._height
        )
        self.assertIs(foo.data, self.item.image)

        foo = self.scaling.scale(
            "image",
            width=self.item.image._width,
            height=self.item.image._height,
            quality=80,
        )
        self.assertIsNot(foo.data, self.item.image)

    def testCreateHighPixelDensityScale(self):
        self.scaling.getHighPixelDensityScales = lambda: [{"scale": 2, "quality": 66}]
        foo = self.scaling.scale("image", width=100, height=80, include_srcset=True)
        self.assertTrue(foo.srcset)
        self.assertEqual(foo.srcset[0]["mimetype"], "image/png")
        self.assertEqual(foo.srcset[0]["height"], 160)
        self.assertEqual(foo.srcset[0]["width"], 160)
        # It is a pre-registered scale, not yet rendered.
        self.assertEqual(foo.srcset[0]["data"], None)
        # Render the scale by pretending to visit its url.
        bar = self.scaling.publishTraverse(self.layer["request"], foo.srcset[0]["uid"])
        assertImage(self, bar.data.data, "PNG", (160, 160))

    def testCreateScaleWithoutData(self):
        item = DummyContent()
        scaling = ImageScaling(item, None)
        foo = scaling.scale("image", width=100, height=80)
        self.assertEqual(foo, None)

    def testCreateHighPixelDensityScaleWithoutData(self):
        item = DummyContent()
        scaling = ImageScaling(item, None)
        scaling.getHighPixelDensityScales = lambda: [{"scale": 2, "quality": 66}]
        foo = scaling.scale("image", width=100, height=80)
        self.assertFalse(hasattr(foo, "srcset"))

    def testGetScaleByName(self):
        self.scaling.available_sizes = {"foo": (60, 60)}
        foo = self.scaling.scale("image", scale="foo")
        self.assertTrue(foo.uid)
        self.assertEqual(foo.mimetype, "image/png")
        self.assertIsInstance(foo.mimetype, str)
        self.assertEqual(foo.data.contentType, "image/png")
        self.assertIsInstance(foo.data.contentType, str)
        self.assertEqual(foo.width, 60)
        self.assertEqual(foo.height, 60)
        assertImage(self, foo.data.data, "PNG", (60, 60))
        expected_url = re.compile(rf"http://nohost/item/@@images/{PAT_UID_SCALE}.png")
        self.assertTrue(expected_url.match(foo.absolute_url()))
        self.assertEqual(foo.url, foo.absolute_url())

        tag = foo.tag()
        base = self.item.absolute_url()
        expected = (
            r'<img src="{}/@@images/({}).(jpeg|gif|png)" '
            r'alt="foo" title="foo" height="(\d+)" width="(\d+)" />'.format(
                base, PAT_UID_SCALE
            )
        )
        groups = re.match(expected, tag).groups()
        self.assertTrue(groups, tag)

    def testGetHighPixelDensityScaleByName(self):
        self.scaling.getHighPixelDensityScales = lambda: [{"scale": 2, "quality": 66}]
        self.scaling.available_sizes = {"foo": (60, 60)}
        foo = self.scaling.scale("image", scale="foo", include_srcset=True)
        self.assertTrue(foo.srcset)
        self.assertEqual(foo.srcset[0]["mimetype"], "image/png")
        self.assertEqual(foo.srcset[0]["width"], 120)
        self.assertEqual(foo.srcset[0]["height"], 120)

        # It is a pre-registered scale, not yet rendered.
        self.assertEqual(foo.srcset[0]["data"], None)
        # Render the scale by pretending to visit its url.
        bar = self.scaling.publishTraverse(self.layer["request"], foo.srcset[0]["uid"])
        assertImage(self, bar.data.data, "PNG", (120, 120))

        tag = foo.tag()
        base = self.item.absolute_url()
        expected = (
            rf'<img src="{base}'
            + rf"/@@images/({PAT_UID_SCALE})"
            + r'.(jpeg|gif|png)" '
            r'alt="foo" title="foo" height="(\d+)" width="(\d+)" '
            r'srcset="http://nohost/item/@@images/({})'.format(PAT_UID_SCALE)
            + r".(jpeg|gif|png)"
            r' 2x" />'
        )
        groups = re.match(expected, tag).groups()
        self.assertTrue(groups, tag)

    def testGetRetinaScaleByWidthAndHeight(self):
        self.scaling.getHighPixelDensityScales = lambda: [{"scale": 2, "quality": 66}]
        foo = self.scaling.scale("image", width=60, height=60, include_srcset=True)
        self.assertTrue(foo.srcset)
        self.assertEqual(foo.srcset[0]["mimetype"], "image/png")
        self.assertEqual(foo.srcset[0]["width"], 120)
        self.assertEqual(foo.srcset[0]["height"], 120)

        # It is a pre-registered scale, not yet rendered.
        self.assertEqual(foo.srcset[0]["data"], None)
        # Render the scale by pretending to visit its url.
        bar = self.scaling.publishTraverse(self.layer["request"], foo.srcset[0]["uid"])
        assertImage(self, bar.data.data, "PNG", (120, 120))

        tag = foo.tag()
        base = self.item.absolute_url()
        expected = (
            rf'<img src="{base}'
            + rf"/@@images/({PAT_UID_SCALE})"
            + r'.(jpeg|gif|png)" '
            r'alt="foo" title="foo" height="(\d+)" width="(\d+)" '
            r'srcset="http://nohost/item/@@images/({})'.format(PAT_UID_SCALE)
            + r".(jpeg|gif|png)"
            r' 2x" />'
        )
        groups = re.match(expected, tag).groups()
        self.assertTrue(groups, tag)

    def testGetRetinaScaleByWidthOnly(self):
        self.scaling.getHighPixelDensityScales = lambda: [{"scale": 2, "quality": 66}]
        foo = self.scaling.scale("image", width=60, include_srcset=True)
        self.assertTrue(foo.srcset)
        self.assertEqual(foo.srcset[0]["mimetype"], "image/png")
        self.assertEqual(foo.srcset[0]["width"], 120)
        self.assertEqual(foo.srcset[0]["height"], 120)
        # It is a pre-registered scale, not yet rendered.
        self.assertEqual(foo.srcset[0]["data"], None)
        # Render the scale by pretending to visit its url.
        bar = self.scaling.publishTraverse(self.layer["request"], foo.srcset[0]["uid"])
        assertImage(self, bar.data.data, "PNG", (120, 120))

        tag = foo.tag()
        base = self.item.absolute_url()
        expected = (
            rf'<img src="{base}'
            + rf"/@@images/({PAT_UID_SCALE})"
            + r'.(jpeg|gif|png)" '
            r'alt="foo" title="foo" height="(\d+)" width="(\d+)" '
            r'srcset="http://nohost/item/@@images/({})'.format(PAT_UID_SCALE)
            + r".(jpeg|gif|png)"
            r' 2x" />'
        )
        groups = re.match(expected, tag).groups()
        self.assertTrue(groups, tag)

    def testGetRetinaScaleByHeightOnly(self):
        self.scaling.getHighPixelDensityScales = lambda: [{"scale": 2, "quality": 66}]
        foo = self.scaling.scale("image", height=60, include_srcset=True)
        self.assertTrue(foo.srcset)
        self.assertEqual(foo.srcset[0]["mimetype"], "image/png")
        self.assertEqual(foo.srcset[0]["width"], 120)
        self.assertEqual(foo.srcset[0]["height"], 120)
        # It is a pre-registered scale, not yet rendered.
        self.assertEqual(foo.srcset[0]["data"], None)
        # Render the scale by pretending to visit its url.
        bar = self.scaling.publishTraverse(self.layer["request"], foo.srcset[0]["uid"])
        assertImage(self, bar.data.data, "PNG", (120, 120))

        tag = foo.tag()
        base = self.item.absolute_url()
        expected = (
            rf'<img src="{base}'
            + rf"/@@images/({PAT_UID_SCALE})"
            + r'.(jpeg|gif|png)" '
            r'alt="foo" title="foo" height="(\d+)" width="(\d+)" '
            r'srcset="http://nohost/item/@@images/({})'.format(PAT_UID_SCALE)
            + r".(jpeg|gif|png)"
            r' 2x" />'
        )
        groups = re.match(expected, tag).groups()
        self.assertTrue(groups, tag)

    @patch.object(
        plone.namedfile.scaling,
        "get_picture_variants",
        new=patch_Img2PictureTag_picture_variants,
        spec=True,
    )
    @patch.object(
        plone.namedfile.picture,
        "get_allowed_scales",
        new=patch_Img2PictureTag_allowed_scales,
        spec=True,
    )
    @patch.object(plone.namedfile.picture, "uuidToObject", spec=True)
    def testGetPictureTagByName(self, mock_uuid_to_object):
        ImageScaling._sizes = patch_Img2PictureTag_allowed_scales()
        mock_uuid_to_object.return_value = self.item
        tag = self.scaling.picture("image", picture_variant="medium")
        expected = """<picture>
 <source srcset="http://nohost/item/@@images/image-600-....png 600w,
http://nohost/item/@@images/image-400-....png 400w,
http://nohost/item/@@images/image-800-....png 800w,
http://nohost/item/@@images/image-1000-....png 1000w,
http://nohost/item/@@images/image-1200-....png 1200w"/>
 <img...src="http://nohost/item/@@images/image-600-....png".../>
</picture>"""
        self.assertTrue(_ellipsis_match(expected, tag.strip()))

        # The exact placement of the img tag attributes can differ, especially
        # with different beautifulsoup versions.
        # So check here that all attributes are present.
        self.assertIn('height="200"', tag)
        self.assertIn('loading="lazy"', tag)
        self.assertIn('title="foo"', tag)
        self.assertIn('width="200"', tag)

    @patch.object(
        plone.namedfile.scaling,
        "get_picture_variants",
        new=patch_Img2PictureTag_picture_variants,
        spec=True,
    )
    @patch.object(
        plone.namedfile.picture,
        "get_allowed_scales",
        new=patch_Img2PictureTag_allowed_scales,
        spec=True,
    )
    @patch.object(plone.namedfile.picture, "uuidToObject", spec=True)
    def testGetPictureTagWithAltAndTitle(self, mock_uuid_to_object):
        ImageScaling._sizes = patch_Img2PictureTag_allowed_scales()
        mock_uuid_to_object.return_value = self.item
        tag = self.scaling.picture(
            "image",
            picture_variant="medium",
            alt="Alternative text",
            title="Custom title",
        )
        base = self.item.absolute_url()
        expected = f"""<picture>
 <source srcset="{base}/@@images/image-600-....png 600w,
{base}/@@images/image-400-....png 400w,
{base}/@@images/image-800-....png 800w,
{base}/@@images/image-1000-....png 1000w,
{base}/@@images/image-1200-....png 1200w"/>
 <img...src="{base}/@@images/image-600-....png".../>
</picture>"""
        self.assertTrue(_ellipsis_match(expected, tag.strip()))

        # The exact placement of the img tag attributes can differ, especially
        # with different beautifulsoup versions.
        # So check here that all attributes are present.
        self.assertIn('alt="Alternative text"', tag)
        self.assertIn('height="200"', tag)
        self.assertIn('loading="lazy"', tag)
        self.assertIn('title="Custom title"', tag)
        self.assertIn('width="200"', tag)

    @patch.object(
        plone.namedfile.scaling,
        "get_picture_variants",
        new=patch_Img2PictureTag_empty_picture_variants,
        spec=True,
    )
    @patch.object(
        plone.namedfile.picture,
        "get_allowed_scales",
        new=patch_Img2PictureTag_allowed_scales,
        spec=True,
    )
    @patch.object(plone.namedfile.picture, "uuidToObject", spec=True)
    def testGetPictureTagWithoutAnyVariants(self, mock_uuid_to_object):
        ImageScaling._sizes = patch_Img2PictureTag_allowed_scales()
        mock_uuid_to_object.return_value = self.item
        tag = self.scaling.picture("image", picture_variant="medium")
        expected = """<img...src="http://nohost/item/@@images/image-0-....png".../>"""
        self.assertTrue(_ellipsis_match(expected, tag.strip()))

        # The exact placement of the img tag attributes can differ, especially
        # with different beautifulsoup versions.
        # So check here that all attributes are present.
        self.assertIn('height="200"', tag)
        self.assertIn('title="foo"', tag)
        self.assertIn('width="200"', tag)

    def testGetUnknownScale(self):
        foo = self.scaling.scale("image", scale="foo?")
        self.assertEqual(foo, None)

    def testScaleInvalidation(self):
        dt = self.item.modified()

        # Test that different parameters give different scale
        self.item.modified = lambda: dt
        scale1a = self.scaling.scale("image", width=100, height=80)
        scale2a = self.scaling.scale("image", width=80, height=60)
        self.assertNotEqual(scale1a.data, scale2a.data)

        # Test that bare object modification does not invalidate scales
        self.item.modified = lambda: dt + 1
        scale1b = self.scaling.scale("image", width=100, height=80)
        scale2b = self.scaling.scale("image", width=80, height=60)
        self.assertNotEqual(scale1b.data, scale2b.data)
        self.assertEqual(scale1a.data, scale1b.data)
        self.assertEqual(scale2a.data, scale2b.data)

        # Test changing _p_mtime on the field no longer invalidates a scale
        self.item.image._p_mtime = (dt + 1).millis()
        scale1b = self.scaling.scale("image", width=100, height=80)
        scale2b = self.scaling.scale("image", width=80, height=60)
        self.assertNotEqual(scale1b.data, scale2b.data)
        self.assertEqual(scale1a.data, scale1b.data)
        self.assertEqual(scale2a.data, scale2b.data)

        # Test changing the fields modified timestamp invalidates scales
        self.item.image._modified = (dt + 1).millis()
        scale1c = self.scaling.scale("image", width=100, height=80)
        scale2c = self.scaling.scale("image", width=80, height=60)
        self.assertNotEqual(scale1c.data, scale2c.data)
        self.assertNotEqual(scale1a.data, scale1c.data, "scale not updated?")
        self.assertNotEqual(scale2a.data, scale2c.data, "scale not updated?")

    def testFallBackToDatabaseModifiedTimeStamp(self):
        dt = self.item.modified()
        scale_a = self.scaling.scale("image", width=100, height=80)

        delattr(self.item.image, "_modified")

        # Since there is no _modified timestamp, _p_mtime is the fallback.
        self.item.image._p_mtime = (dt + 1).millis()
        scale_b = self.scaling.scale("image", width=100, height=80)
        self.assertNotEqual(scale_a.data, scale_b.data)

    def testCustomSizeChange(self):
        # set custom image sizes & view a scale
        self.scaling.available_sizes = {"foo": (23, 23)}
        foo = self.scaling.scale("image", scale="foo")
        self.assertEqual(foo.width, 23)
        self.assertEqual(foo.height, 23)
        # now let's update the scale dimensions, after which the scale
        # shouldn't be the same...
        self.scaling.available_sizes = {"foo": (42, 42)}
        foo = self.scaling.scale("image", scale="foo")
        self.assertEqual(foo.width, 42)
        self.assertEqual(foo.height, 42)

    def testDefaultAvailableSizes(self):
        # by default, no named scales are configured
        self.assertEqual(self.scaling.available_sizes, {})

    def testCustomAvailableSizes(self):
        # a callable can be used to look up the available sizes
        def custom_available_sizes():
            return {"bar": (10, 10)}

        sm = getSiteManager()
        sm.registerUtility(component=custom_available_sizes, provided=IAvailableSizes)
        self.assertEqual(self.scaling.available_sizes, {"bar": (10, 10)})
        sm.unregisterUtility(provided=IAvailableSizes)
        # for testing purposes, the sizes may also be set directly on
        # the scaling adapter
        self.scaling.available_sizes = {"qux": (12, 12)}
        self.assertEqual(self.scaling.available_sizes, {"qux": (12, 12)})

    def testGuardedAccess(self):
        # make sure it's not possible to access scales of forbidden images
        self.item.__allow_access_to_unprotected_subobjects__ = 0
        self.assertRaises(Unauthorized, self.scaling.guarded_orig_image, "image")
        self.item.__allow_access_to_unprotected_subobjects__ = 1

    def testGetAvailableSizes(self):
        self.scaling.available_sizes = {"foo": (60, 60)}
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("once")
            self.assertEqual(
                self.scaling.getAvailableSizes(),
                {"foo": (60, 60)},
            )
            self.assertEqual(len(w), 1)
            self.assertIs(w[0].category, DeprecationWarning)
            self.assertIn(
                "use property available_sizes instead",
                str(w[0].message),
            )
            self.assertEqual(
                self.scaling.getAvailableSizes("image"),
                {"foo": (60, 60)},
            )
            self.assertEqual(len(w), 2)
            self.assertIs(w[1].category, DeprecationWarning)
            self.assertIn(
                "fieldname was passed to deprecated getAvailableSizes, but "
                "will be ignored.",
                str(w[1].message),
            )

    def testGetImageSize(self):
        assert self.scaling.getImageSize("image") == (200, 200)

    def testGetOriginalScaleTag(self):
        tag = self.scaling.tag("image")
        base = self.item.absolute_url()
        expected = (
            r'<img src="{}/@@images/({}).(jpeg|gif|png)" '
            r'alt="foo" title="foo" height="(\d+)" width="(\d+)" />'.format(
                base, PAT_UID_SCALE
            )
        )
        self.assertTrue(re.match(expected, tag).groups())

    def testScaleOnItemWithNonASCIITitle(self):
        self.item.title = "ü"
        tag = self.scaling.tag("image")
        base = self.item.absolute_url()
        expected = (
            r'<img src="{}/@@images/({}).(jpeg|gif|png)" '
            r'alt="\xfc" title="\xfc" height="(\d+)" width="(\d+)" />'.format(
                base, PAT_UID_SCALE
            )
        )
        self.assertTrue(re.match(expected, tag).groups())

    def testScaleOnItemWithUnicodeTitle(self):
        self.item.Title = lambda: "ü"
        tag = self.scaling.tag("image")
        base = self.item.absolute_url()
        expected = (
            r'<img src="{}/@@images/({}).(jpeg|gif|png)" '
            r'alt="\xfc" title="\xfc" height="(\d+)" width="(\d+)" />'.format(
                base, PAT_UID_SCALE
            )
        )
        self.assertTrue(re.match(expected, tag).groups())

    def testScaledJpegImageQuality(self):
        """Test image quality setting for jpeg images.

        Image quality is not available for PNG images.
        """
        dt = DateTime()
        data = getFile("image.jpg")
        item = DummyContent()
        item.image = NamedImage(data, "image/png", "image.jpg")
        item.image._modified = dt.millis()
        scaling = ImageScaling(item, None)

        # scale an image, record its size
        foo = scaling.scale("image", width=100, height=80)
        size_foo = foo.data.getSize()
        # Let's pretend the imaging control panel sets the scaling quality to
        # "really sloppy"
        gsm = getGlobalSiteManager()
        qualitySupplier = DummyQualitySupplier()
        gsm.registerUtility(qualitySupplier.getQuality, IScaledImageQuality)
        item.image._modified = (dt + 1).millis()
        # now scale again
        bar = scaling.scale("image", width=100, height=80)
        size_bar = bar.data.getSize()
        # first one should be bigger
        self.assertTrue(size_foo > size_bar)

    def testOversizedHighPixelDensityScale(self):
        orig_size = max(self.scaling.getImageSize("image"))
        scale_size = orig_size / 2
        self.scaling.getHighPixelDensityScales = lambda: [
            {"scale": 2, "quality": 66},
            {"scale": 3, "quality": 66},
        ]
        foo = self.scaling.scale(
            "image", width=scale_size, height=scale_size, include_srcset=True
        )
        self.assertEqual(len(foo.srcset), 1)
        self.assertEqual(foo.srcset[0]["scale"], 2)


class ImageTraverseTests(unittest.TestCase):

    layer = PLONE_NAMEDFILE_FUNCTIONAL_TESTING

    def setUp(self):
        self.app = self.layer["app"]
        data = getFile("image.png")
        item = DummyContent()
        item.image = NamedImage(data, "image/png", "image.png")
        self.app._setOb("item", item)
        self.item = self.app.item
        self._orig_sizes = ImageScaling._sizes

    def tearDown(self):
        ImageScaling._sizes = self._orig_sizes

    def traverse(self, path=""):
        view = self.item.unrestrictedTraverse("@@images")
        stack = path.split("/")
        name = stack.pop(0)
        static_traverser = view.traverse(name, stack)
        scale = stack.pop(0)
        tag = static_traverser.traverse(scale, stack)
        base = self.item.absolute_url()
        expected = (
            r'<img src="{0}/@@images/([0-9a-z]*-[0-9]*-[0-9a-f]{{32}}).(jpeg|gif|png)" '
            r'alt="foo" title="foo" height="(\d+)" width="(\d+)" />'.format(
                base,
            )
        )
        groups = re.match(expected, tag).groups()
        self.assertTrue(groups, tag)
        uid, ext, height, width = groups
        return uid, ext, int(width), int(height)

    def testImageThumb(self):
        ImageScaling._sizes = {"thumb": (128, 128)}
        uid, ext, width, height = self.traverse("image/thumb")
        self.assertEqual((width, height), ImageScaling._sizes["thumb"])
        self.assertEqual(ext, "png")

    def testCustomSizes(self):
        # set custom image sizes
        ImageScaling._sizes = {"foo": (23, 23)}
        # make sure traversing works with the new sizes
        uid, ext, width, height = self.traverse("image/foo")
        self.assertEqual(width, 23)
        self.assertEqual(height, 23)

    def testScaleInvalidation(self):
        # first view the thumbnail of the original image
        ImageScaling._sizes = {"thumb": (128, 128)}
        uid1, ext, width1, height1 = self.traverse("image/thumb")
        wait_to_ensure_modified()
        # now upload a new one and make sure the thumbnail has changed
        data = getFile("image.jpg")
        self.item.image = NamedImage(data, "image/jpeg", "image.jpg")
        uid2, ext, width2, height2 = self.traverse("image/thumb")
        self.assertNotEqual(uid1, uid2, "thumb not updated?")
        # the height also differs as the original image had a size of 200, 200
        # whereas the updated one has 500, 200...
        self.assertEqual(width1, width2)
        self.assertNotEqual(height1, height2)

    def testCustomSizeChange(self):
        # set custom image sizes & view a scale
        ImageScaling._sizes = {"foo": (23, 23)}
        uid1, ext, width, height = self.traverse("image/foo")
        self.assertEqual(width, 23)
        self.assertEqual(height, 23)
        # now let's update the scale dimensions, after which the scale
        # should also have been updated...
        ImageScaling._sizes = {"foo": (42, 42)}
        uid2, ext, width, height = self.traverse("image/foo")
        self.assertEqual(width, 42)
        self.assertEqual(height, 42)
        self.assertNotEqual(uid1, uid2, "scale not updated?")

    def testGuardedAccess(self):
        # make sure it's not possible to access scales of forbidden images
        self.item.__allow_access_to_unprotected_subobjects__ = 0
        ImageScaling._sizes = {"foo": (42, 42)}
        self.assertRaises(Unauthorized, self.traverse, "image/foo")
        self.item.__allow_access_to_unprotected_subobjects__ = 1


class StorageTests(unittest.TestCase):
    """Test the scale storage.

    Especially test that we get an adapter for IImageScaleStorage
    instead of defaulting to plone.scale.storage.AnnotationStorage.
    """

    layer = PLONE_NAMEDFILE_INTEGRATION_TESTING

    def setUp(self):
        sm = getSiteManager()
        sm.registerAdapter(FakeImageScaleStorage, (FakeImage, Interface))

    def tearDown(self):
        sm = getSiteManager()
        sm.unregisterAdapter(FakeImageScaleStorage, (FakeImage, Interface))

    def test_original(self):
        item = FakeImage("abcdef", "jpeg")
        scaling = ImageScaling(item, None)
        original = scaling.scale("image")
        self.assertEqual(original.data.value, "abcdef")
        self.assertIs(original.data, item)
        self.assertEqual(original.mimetype, "image/jpeg")
        self.assertIsInstance(original.mimetype, str)
        self.assertEqual(original.data.contentType, "image/jpeg")
        self.assertIsInstance(original.data.contentType, str)
        self.assertEqual(original.width, 6)
        self.assertEqual(original.height, 4)

        # Try the tag
        self.assertEqual(
            scaling.tag("image"),
            '<img src="http://fake.image/@@images/image.jpeg" alt="Image Title" title="Image Title" height="4" width="6" />',
        )

        # Try to access the item via uid.
        # This should fail, because we only have the original, not a scale.
        with self.assertRaises(NotFound):
            scaling.publishTraverse(self.layer["request"], "uid-0")

    def test_width(self):
        # For FakeImage scales, the width changes the mimetype/format.
        item = FakeImage("abcdef", "jpeg")
        scaling = ImageScaling(item, None)
        scale = scaling.scale("image", width=2)
        self.assertEqual(scale.data.value, "abcdef")
        self.assertIsInstance(scale.data, FakeImage)
        self.assertEqual(scale.mimetype, "image/jp")
        self.assertEqual(scale.data.contentType, "image/jp")
        self.assertEqual(scale.width, 6)
        self.assertEqual(scale.height, 2)

        # Ask for the same scale and you get the same FakeImage.
        scale2 = scaling.scale("image", width=2)
        self.assertIs(scale.data, scale2.data)

        # Try the tag.  It should have a uid.
        self.assertEqual(
            scaling.tag("image", width=2),
            '<img src="http://fake.image/@@images/uid-0.jp" alt="Image Title" title="Image Title" height="2" width="6" />',
        )

        # Access the item via uid.
        scale_uid = scaling.publishTraverse(self.layer["request"], "uid-0")
        self.assertEqual(scale_uid.data.uid, "uid-0")
        self.assertEqual(scale_uid.data.info["uid"], "uid-0")
        self.assertIs(scale_uid.data, scale2.data)

    def test_height(self):
        # For FakeImage scales, the height changes the value.
        item = FakeImage("abcdef", "jpeg")
        scaling = ImageScaling(item, None)
        scale = scaling.scale("image", height=3)
        self.assertEqual(scale.data.value, "abc")
        self.assertIsInstance(scale.data, FakeImage)
        self.assertEqual(scale.mimetype, "image/jpeg")
        self.assertEqual(scale.data.contentType, "image/jpeg")
        self.assertEqual(scale.width, 3)
        self.assertEqual(scale.height, 4)

        # Ask for the same scale and you get the same FakeImage.
        scale2 = scaling.scale("image", height=3)
        self.assertIs(scale.data, scale2.data)

        # Try the tag.  It should have a uid.
        self.assertEqual(
            scaling.tag("image", height=3),
            '<img src="http://fake.image/@@images/uid-0.jpeg" alt="Image Title" title="Image Title" height="4" width="3" />',
        )

        # Access the item via uid.
        scale_uid = scaling.publishTraverse(self.layer["request"], "uid-0")
        self.assertEqual(scale_uid.data.uid, "uid-0")
        self.assertEqual(scale_uid.data.info["uid"], "uid-0")
        self.assertIs(scale_uid.data, scale2.data)

    def test_title(self):
        # Test that a custom title property on an ImageScale class is used.
        item = FakeImage("abcdef", "jpeg")
        scaling = ImageScaling(item, None)
        scaling._scale_view_class = TitleImageScale
        self.assertEqual(
            scaling.tag("image"),
            '<img src="http://fake.image/@@images/image.jpeg" alt="title from class" title="title from class" height="4" width="6" />'
        )
        self.assertEqual(
            scaling.tag("image", alt="own alt"),
            '<img src="http://fake.image/@@images/image.jpeg" alt="own alt" title="title from class" height="4" width="6" />'
        )
        self.assertEqual(
            scaling.tag("image", title="own title"),
            '<img src="http://fake.image/@@images/image.jpeg" alt="title from class" title="own title" height="4" width="6" />'
        )
        self.assertEqual(
            scaling.tag("image", alt="own alt", title="own title"),
            '<img src="http://fake.image/@@images/image.jpeg" alt="own alt" title="own title" height="4" width="6" />'
        )


class Img2PictureTagTests(unittest.TestCase):
    """Low level tests for Img2PictureTag."""

    def _makeOne(self):
        return plone.namedfile.picture.Img2PictureTag()

    def test_update_src_scale(self):
        update_src_scale = self._makeOne().update_src_scale
        self.assertEqual(
            update_src_scale("foo/fieldname/old", "new"),
            "foo/fieldname/new"
        )
        self.assertEqual(
            update_src_scale("@@images/fieldname/old", "mini"),
            "@@images/fieldname/mini"
        )
        self.assertEqual(
            update_src_scale("@@images/fieldname", "preview"),
            "@@images/fieldname/preview"
        )
        self.assertEqual(
            update_src_scale(
                "photo.jpg/@@images/image-1200-4a03b0a8227d28737f5d9e3e481bdbd6.jpeg",
                "teaser"),
            "photo.jpg/@@images/image/teaser",
        )


def test_suite():
    from unittest import defaultTestLoader

    return defaultTestLoader.loadTestsFromName(__name__)



from collections.abc import Iterable
from io import BytesIO
from io import FileIO
from logging import getLogger
from plone.namedfile.interfaces import IBlobby
from plone.namedfile.utils.jpeg_utils import process_jpeg
from plone.namedfile.utils.png_utils import process_png
from plone.namedfile.utils.svg_utils import process_svg
from plone.registry.interfaces import IRegistry
from urllib.parse import quote
from zope.component import queryUtility
from zope.deprecation import deprecate
from zope.interface import implementer
from ZPublisher.Iterators import IStreamIterator

import mimetypes
import piexif
import PIL.Image
import re
import struct

# image-scaling
QUALITY_DEFAULT = 88
pattern = re.compile(r'^(.*)\s+(\d+)\s*:\s*(\d+)$')

log = getLogger(__name__)

try:
    from plone.base.interfaces.controlpanel import IImagingSchema
except ImportError:
    from Products.CMFPlone.interfaces.controlpanel import IImagingSchema

try:
    # Zope 5.8.6+
    from OFS.Image import extract_media_type
except ImportError:

    def extract_media_type(content_type):
        """extract the proper media type from *content_type*.

        Ignore parameters and whitespace and normalize to lower case.
        See https://github.com/zopefoundation/Zope/pull/1167
        """
        if not content_type:
            return content_type
        # ignore parameters
        content_type = content_type.split(";", 1)[0]
        # ignore whitespace
        content_type = "".join(content_type.split())
        # normalize to lowercase
        return content_type.lower()


@implementer(IStreamIterator)
class filestream_range_iterator(Iterable):
    """
    A class that mimics FileIO and implements an iterator that returns a
    fixed-sized sequence of bytes. Beginning from `start` to `end`.

    BBB: due to a possible bug in Zope>4, <=4.1.3, couldn't be subclass of FileIO
         as Iterators.filestream_iterator
    """

    def __init__(
        self, name, mode="rb", bufsize=-1, streamsize=1 << 16, start=0, end=None
    ):
        self._io = FileIO(name, mode=mode)
        self.streamsize = streamsize
        self.start = start
        self.end = end
        self._io.seek(start, 0)

    def __iter__(self):
        if self._io.closed:
            raise ValueError("I/O operation on closed file.")
        return self

    def __next__(self):
        if self.end is None:
            bytes = self.streamsize
        else:
            bytes = max(min(self.end - self._io.tell(), self.streamsize), 0)
        data = self._io.read(bytes)
        if not data:
            raise StopIteration
        return data

    next = __next__

    def close(self):
        self._io.close()

    # BBB: is it necessary to implement __len__ ?
    # def __len__(self)

    def read(self, size=-1):
        return self._io.read(size)


def safe_basename(filename):
    """Get the basename of the given filename, regardless of which platform
    (Windows or Unix) it originated from.
    """
    fslice = (
        max(
            filename.rfind("/"),
            filename.rfind("\\"),
            filename.rfind(":"),
        )
        + 1
    )
    return filename[fslice:]


def get_contenttype(file=None, filename=None, default="application/octet-stream"):
    """Get the MIME content type of the given file and/or filename.

    Note: depending on your use case, you may want to call 'extract_media_type'
    on the result.
    """

    file_type = getattr(file, "contentType", None)
    if file_type:
        return file_type

    filename = getattr(file, "filename", filename)
    if filename:
        return mimetypes.guess_type(filename, strict=False)[0] or default

    return default


def set_headers(file, response, filename=None, canonical=None):
    """Set response headers for the given file. If filename is given, set
    the Content-Disposition to attachment.
    """

    contenttype = get_contenttype(file)

    response.setHeader("Content-Type", contenttype)
    response.setHeader("Content-Length", file.getSize())
    response.setHeader("Accept-Ranges", "bytes")

    if filename is not None:
        if not isinstance(filename, str):
            filename = str(filename, "utf-8", errors="ignore")
        filename = quote(filename.encode("utf8"))
        response.setHeader(
            "Content-Disposition", f"attachment; filename*=UTF-8''{filename}"
        )

    if canonical is not None:
        response.setHeader("Link", f'<{quote(canonical, safe="/:&?=@")}>; rel="canonical"')

def stream_data(file, start=0, end=None):
    """Return the given file as a stream if possible."""
    if IBlobby.providedBy(file):
        if file._blob._p_blob_uncommitted:
            return file.data[start:end]
        return filestream_range_iterator(
            file._blob.committed(), "rb", start=start, end=end
        )
    return file.data[start:end]


def _ensure_data(image):
    data = None
    if getattr(image, "read", None):
        data = image.read()
        image.seek(0)
    else:
        data = image
    return bytes(data)


def getImageInfo(data):
    data = _ensure_data(data)
    size = len(data)
    height = -1
    width = -1
    content_type = ""

    if (size >= 10) and data[:6] in (b"GIF87a", b"GIF89a"):
        # handle GIFs
        content_type = "image/gif"
        w, h = struct.unpack("<HH", data[6:10])
        width = int(w)
        height = int(h)

    elif data[:8] == b"\211PNG\r\n\032\n":
        # handle PNG
        content_type, width, height = process_png(data)

    elif data[:2] == b"\377\330":
        # handle JPEGs
        content_type, width, height = process_jpeg(data)

    elif (size >= 30) and data.startswith(b"BM"):
        # handle BMPs
        kind = struct.unpack("<H", data[14:16])[0]
        if kind == 40:  # Windows 3.x bitmap
            content_type = "image/x-ms-bmp"
            width, height = struct.unpack("<LL", data[18:26])

    elif size and b"http://www.w3.org/2000/svg" in data:
        # handle SVGs
        content_type, width, height = process_svg(data)

    else:
        # Use PIL / Pillow to determ Image Information
        try:
            img = PIL.Image.open(BytesIO(data))
            width, height = img.size
            content_type = PIL.Image.MIME[img.format]
        except Exception:
            # TODO: determ wich error really happens
            # Should happen if data is to short --> first_bytes
            # happens also if data is an svg or another special format.
            log.warning(
                "PIL can not recognize the image. "
                "Image is probably broken or of a non-supported format."
            )

    log.debug(
        "Image Info (Type: %s, Width: %s, Height: %s)", content_type, width, height
    )
    return content_type, width, height


def get_exif(image):
    #
    exif_data = None
    image_data = _ensure_data(image)

    content_type, width, height = getImageInfo(image_data)
    if content_type in ["image/jpeg", "image/tiff"]:
        # Only this two Image Types could have Exif informations
        # see http://www.cipa.jp/std/documents/e/DC-008-2012_E.pdf
        try:
            exif_data = piexif.load(image_data)
        except Exception as e:
            # TODO: determ wich error really happens
            # Should happen if data is to short --> first_bytes
            log.warn(e)
            exif_data = exif_data = {
                "0th": {
                    piexif.ImageIFD.XResolution: (width, 1),
                    piexif.ImageIFD.YResolution: (height, 1),
                }
            }
    return exif_data


def rotate_image(image_data, method=None, REQUEST=None):
    """Rotate Image if it has Exif Orientation Informations other than 1.

    Do not use PIL.Image.rotate function as this did not transpose the image,
    rotate keeps the image width and height and rotates the image around a
    central point. PIL.Image.transpose also changes Image Orientation.
    """
    orientation = 1  # if not set assume correct orrinetation --> 1
    data = _ensure_data(image_data)
    img = PIL.Image.open(BytesIO(data))

    exif_data = None
    if "exif" in img.info:
        try:
            exif_data = piexif.load(img.info["exif"])
        except ValueError:
            log.warn("Exif information currupt")
            pass
        if exif_data and piexif.ImageIFD.Orientation in exif_data["0th"]:
            orientation = exif_data["0th"][piexif.ImageIFD.Orientation]
        if exif_data and (
            not exif_data["0th"].get(piexif.ImageIFD.XResolution)
            or not exif_data["0th"].get(piexif.ImageIFD.YResolution)
        ):
            exif_data["0th"][piexif.ImageIFD.XResolution] = (img.width, 1)
            exif_data["0th"][piexif.ImageIFD.YResolution] = (img.height, 1)
    if exif_data is None:
        width, height = img.size
        exif_data = {
            "0th": {
                piexif.ImageIFD.XResolution: (width, 1),
                piexif.ImageIFD.YResolution: (height, 1),
            }
        }

    if method is not None:
        orientation = method

    log.debug("Rotate image with input orientation: %s", orientation)

    fmt = img.format
    if orientation == 1:  # not transform necessary
        # img = img
        pass
    elif orientation == 2:
        img = img.transpose(PIL.Image.FLIP_LEFT_RIGHT)
    elif orientation == 3:
        img = img.transpose(PIL.Image.ROTATE_180)
    elif orientation == 4:
        img = img.transpose(PIL.Image.ROTATE_180).transpose(PIL.Image.FLIP_LEFT_RIGHT)
    elif orientation == 5:
        img = img.transpose(PIL.Image.ROTATE_270).transpose(PIL.Image.FLIP_LEFT_RIGHT)
    elif orientation == 6:
        img = img.transpose(PIL.Image.ROTATE_270)
    elif orientation == 7:
        img = img.transpose(PIL.Image.ROTATE_90).transpose(PIL.Image.FLIP_LEFT_RIGHT)
    elif orientation == 8:
        img = img.transpose(PIL.Image.ROTATE_90)

    if orientation in [5, 6, 7, 8]:
        if (
            exif_data["0th"][piexif.ImageIFD.XResolution]
            and exif_data["0th"][piexif.ImageIFD.YResolution]
        ):
            (
                exif_data["0th"][piexif.ImageIFD.XResolution],
                exif_data["0th"][piexif.ImageIFD.YResolution],
            ) = (
                exif_data["0th"][piexif.ImageIFD.YResolution],
                exif_data["0th"][piexif.ImageIFD.XResolution],
            )
        else:
            (
                exif_data["0th"][piexif.ImageIFD.XResolution],
                exif_data["0th"][piexif.ImageIFD.YResolution],
            ) = (img.width, 1), (img.height, 1)

    # set orientation to normal
    exif_data["0th"][piexif.ImageIFD.Orientation] = 1

    try:
        exif_bytes = piexif.dump(exif_data)
    except Exception as e:
        log.warn(e)
        del exif_data["Exif"][piexif.ExifIFD.SceneType]
        # This Element piexif.ExifIFD.SceneType cause error on dump
        exif_bytes = piexif.dump(exif_data)

    output_image_data = BytesIO()
    img.save(output_image_data, format=fmt, exif=exif_bytes)
    width, height = img.size
    return output_image_data.getvalue(), width, height, exif_data


@deprecate("use getHighPixelDensityScales instead")
def getRetinaScales():
    return getHighPixelDensityScales()


def getHighPixelDensityScales():
    registry = queryUtility(IRegistry)
    if not registry:
        return []
    settings = registry.forInterface(IImagingSchema, prefix="plone", check=False)
    if settings.highpixeldensity_scales == "2x":
        return [
            {"scale": 2, "quality": settings.quality_2x},
        ]
    if settings.highpixeldensity_scales == "3x":
        return [
            {"scale": 2, "quality": settings.quality_2x},
            {"scale": 3, "quality": settings.quality_3x},
        ]
    return []

def getAllowedSizes():
    registry = queryUtility(IRegistry)
    if not registry:
        return None
    settings = registry.forInterface(
        IImagingSchema, prefix="plone", check=False)
    if not settings.allowed_sizes:
        return None
    sizes = {}
    for line in settings.allowed_sizes:
        line = line.strip()
        if line:
            name, width, height = pattern.match(line).groups()
            name = name.strip().replace(' ', '_')
            sizes[name] = int(width), int(height)
    return sizes


def getQuality():
    registry = queryUtility(IRegistry)
    if registry:
        settings = registry.forInterface(
            IImagingSchema, prefix="plone", check=False)
        return settings.quality or QUALITY_DEFAULT
    return QUALITY_DEFAULT




from zope import schema
from zope.i18nmessageid import MessageFactory
from zope.interface import Interface
from zope.schema.interfaces import IObject


# We don't have translations here, but this allows
_ = MessageFactory("plone")


HAVE_BLOBS = True


class IFile(Interface):

    contentType = schema.NativeStringLine(
        title="Content Type",
        description="The content type identifies the type of data.",
        default="",
        required=False,
        missing_value="",
    )

    data = schema.Bytes(
        title="Data",
        description="The actual content of the object.",
        default=b"",
        missing_value="",
        required=False,
    )

    def getSize():
        """Return the byte-size of the data of the object."""


class IImage(IFile):
    """This interface defines an Image that can be displayed."""

    def getImageSize():
        """Return a tuple (x, y) that describes the dimensions of
        the object.
        """


class IImageScaleTraversable(Interface):
    """Marker for items that should provide access to image scales for named
    image fields via the @@images view.
    """


class IAvailableSizes(Interface):
    """A callable returning a dictionary of scale name => (width, height)"""


class IStableImageScale(Interface):
    """Marker for image scales when accessed with a UID-based URL.
    These can be cached forever using the plone.stableResource ruleset.
    """


class IPluggableBinaryFieldValidation(Interface):
    def __call__(field, value):
        """validates field and value.

        raises zope.schema.ValidationError
        returns None
        """


class IPluggableFileFieldValidation(IPluggableBinaryFieldValidation):
    """pluggable validation for binary File fields"""


class IPluggableImageFieldValidation(IPluggableBinaryFieldValidation):
    """pluggable validation for binary Image fields"""


# Values


class INamed(Interface):
    """An item with a filename"""

    filename = schema.TextLine(title="Filename", required=False, default=None)


class INamedFile(INamed, IFile):
    """A non-BLOB file with a filename"""


class INamedImage(INamed, IImage):
    """A non-BLOB image with a filename"""


# Fields


class INamedField(IObject):
    """Base field type"""


class INamedFileField(INamedField):
    """Field for storing INamedFile objects."""

    accept = schema.Tuple(
        title=_("namedfile_accept_title", default="accept types"),
        description=_(
            "namedfile_accept_description",
            default=(
                'The media types which are allowed for this field. '
                'Unset to allow any type. '
                'Can be any valid identifier for the "accept" attribute of '
                'the HTML file input, like extensions (e.g. ".mp3") or IANA '
                'media types (e.g. "image/webp").'
            ),
        ),
        value_type=schema.TextLine(),
        default=(),
        required=False,
    )


class INamedImageField(INamedField):
    """Field for storing INamedImage objects."""

    accept = schema.Tuple(
        title=_("namedimage_accept_title", default="accept types"),
        description=_(
            "namedimage_accept_description",
            default=(
                'The media types which are allowed for this image field. '
                'The default is to allow any "image/*" content type. '
                'Unset to allow any type. '
                'Can be any valid identifier for the "accept" attribute of '
                'the HTML file input, like extensions (e.g. ".jpg") or IANA '
                'media types (e.g. "image/webp").'
            ),
        ),
        value_type=schema.TextLine(),
        default=("image/*",),
        required=False,
    )


class IStorage(Interface):
    """Store file data"""

    def store(data, blob):
        """Store the data into the blob
        Raises NonStorable if data is not storable.
        """


class NotStorable(Exception):
    """Data is not storable"""


# Values


class IBlobby(Interface):
    """Marker interface for objects that support blobs."""


class INamedBlobFile(INamedFile, IBlobby):
    """A BLOB file with a filename"""


class INamedBlobImage(INamedImage, IBlobby):
    """A BLOB image with a filename"""


# Fields


class INamedBlobFileField(INamedFileField):
    """Field for storing INamedBlobFile objects."""


class INamedBlobImageField(INamedImageField):
    """Field for storing INamedBlobImage objects."""



from plone.namedfile.testing import PLONE_NAMEDFILE_FUNCTIONAL_TESTING
from plone.testing import layered

import doctest
import unittest


TEST_FILES = [
    "usage.rst",
    "handler.rst",
    "marshaler.rst",
    "utils.rst",
]


def test_suite():
    return unittest.TestSuite(
        [
            layered(
                doctest.DocFileSuite(
                    testfile,
                    package="plone.namedfile",
                    optionflags=doctest.ELLIPSIS,
                ),
                PLONE_NAMEDFILE_FUNCTIONAL_TESTING,
            )
            for testfile in TEST_FILES
        ]
    )



from io import BytesIO
from logging import getLogger

import struct


log = getLogger(__name__)


def process_jpeg(data):
    content_type = None
    w = -1
    h = -1
    size = len(data)

    if (size >= 2) and data.startswith(b"\377\330"):  # handle JPEGs
        content_type = "image/jpeg"
        jpeg = BytesIO(data)
        jpeg.read(2)
        b = jpeg.read(1)
        try:
            while b and ord(b) != 0xDA:
                while ord(b) != 0xFF:
                    b = jpeg.read(1)
                while ord(b) == 0xFF:
                    b = jpeg.read(1)
                if ord(b) >= 0xC0 and ord(b) <= 0xC3:
                    jpeg.read(3)
                    h, w = struct.unpack(">HH", jpeg.read(4))
                    break
                else:
                    jpeg.read(int(struct.unpack(">H", jpeg.read(2))[0]) - 2)
                b = jpeg.read(1)
        except struct.error:
            pass
        except ValueError:
            pass
        except TypeError:
            pass

    width = int(w)
    height = int(h)
    return content_type, width, height



from plone.testing import Layer
from plone.testing import publisher
from plone.testing import zca
from plone.testing import zope
from zope.configuration import xmlconfig


class NamedFileTestLayer(Layer):

    defaultBases = (zope.STARTUP, publisher.PUBLISHER_DIRECTIVES)

    def setUp(self):
        zca.pushGlobalRegistry()

        import plone.namedfile

        xmlconfig.file("testing.zcml", plone.namedfile)

    def tearDown(self):
        # Zap the stacked zca context
        zca.popGlobalRegistry()


PLONE_NAMEDFILE_FIXTURE = NamedFileTestLayer()

PLONE_NAMEDFILE_INTEGRATION_TESTING = zope.IntegrationTesting(
    bases=(PLONE_NAMEDFILE_FIXTURE,),
    name="plone.namedfile:NamedFileTestLayerIntegration",
)

PLONE_NAMEDFILE_FUNCTIONAL_TESTING = zope.FunctionalTesting(
    bases=(PLONE_NAMEDFILE_FIXTURE,),
    name="plone.namedfile:NamedFileTestLayerFunctional",
)



from plone.namedfile import field
from plone.namedfile import interfaces
from plone.schemaeditor.fields import FieldFactory
from zope.i18nmessageid.message import MessageFactory


_ = MessageFactory("plone")


class INamedFileField(interfaces.INamedFileField):

    default = field.NamedFile(
        title=interfaces.INamedFileField["default"].title,
        description=interfaces.INamedFileField["default"].description,
        required=False,
    )

    missing_value = field.NamedFile(
        title=interfaces.INamedFileField["missing_value"].title,
        description=interfaces.INamedFileField["missing_value"].description,
        required=False,
    )


class INamedImageField(interfaces.INamedImageField):

    default = field.NamedImage(
        title=interfaces.INamedImageField["default"].title,
        description=interfaces.INamedImageField["default"].description,
        required=False,
    )

    missing_value = field.NamedImage(
        title=interfaces.INamedImageField["missing_value"].title,
        description=interfaces.INamedImageField["missing_value"].description,
        required=False,
    )


NamedFileFactory = FieldFactory(field.NamedFile, _("File Upload"))
NamedImageFactory = FieldFactory(field.NamedImage, _("Image"))


class INamedBlobFileField(interfaces.INamedBlobFileField):

    default = field.NamedBlobFile(
        title=interfaces.INamedBlobFileField["default"].title,
        description=interfaces.INamedBlobFileField["default"].description,
        required=False,
    )

    missing_value = field.NamedBlobFile(
        title=interfaces.INamedBlobFileField["missing_value"].title,
        description=interfaces.INamedBlobFileField["missing_value"].description,
        required=False,
    )


class INamedBlobImageField(interfaces.INamedBlobImageField):

    default = field.NamedBlobImage(
        title=interfaces.INamedBlobImageField["default"].title,
        description=interfaces.INamedBlobImageField["default"].description,
        required=False,
    )

    missing_value = field.NamedBlobImage(
        title=interfaces.INamedBlobImageField["missing_value"].title,
        description=interfaces.INamedBlobImageField["missing_value"].description,
        required=False,
    )


NamedBlobFileFactory = FieldFactory(field.NamedBlobFile, _("File Upload"))
NamedBlobImageFactory = FieldFactory(field.NamedBlobImage, _("Image"))



from AccessControl.ZopeGuards import guarded_getattr
from Acquisition import aq_base
from DateTime import DateTime
from io import BytesIO
from plone.memoize import ram
from plone.namedfile.file import FILECHUNK_CLASSES
from plone.namedfile.browser import ALLOWED_INLINE_MIMETYPES
from plone.namedfile.browser import DISALLOWED_INLINE_MIMETYPES
from plone.namedfile.browser import USE_DENYLIST
from plone.namedfile.interfaces import IAvailableSizes
from plone.namedfile.interfaces import IStableImageScale
from plone.namedfile.picture import get_picture_variants
from plone.namedfile.picture import Img2PictureTag
from plone.namedfile.utils import extract_media_type
from plone.namedfile.utils import getHighPixelDensityScales
from plone.namedfile.utils import set_headers
from plone.namedfile.utils import stream_data
from plone.protect import PostOnly
from plone.rfc822.interfaces import IPrimaryFieldInfo
from plone.scale.interfaces import IImageScaleFactory
from plone.scale.interfaces import IScaledImageQuality
from plone.scale.scale import scaleImage
from plone.scale.storage import IImageScaleStorage
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone.utils import safe_encode
from Products.Five import BrowserView
from xml.sax.saxutils import quoteattr
from zExceptions import BadRequest
from zExceptions import Unauthorized
from ZODB.blob import BlobFile
from ZODB.POSException import ConflictError
from zope.cachedescriptors.property import Lazy as lazy_property
from zope.component import getMultiAdapter
from zope.component import queryUtility
from zope.deprecation import deprecate
from zope.interface import alsoProvides
from zope.interface import implementer
from zope.publisher.interfaces.browser import IBrowserPublisher
from zope.publisher.interfaces import NotFound
from zope.traversing.interfaces import ITraversable
from zope.traversing.interfaces import TraversalError

import functools
import logging
import warnings


logger = logging.getLogger(__name__)
_marker = object()


def _image_tag_from_values(*values):
    """Turn list of tuples into an img tag.

    Naturally, this should at least contain ("src", "some url").
    """
    parts = ["<img"]
    for k, v in values:
        if v is None:
            continue
        if isinstance(v, int):
            v = str(v)
        elif isinstance(v, bytes):
            v = str(v, "utf8")
        parts.append(f"{k}={quoteattr(v)}")
    parts.append("/>")

    return " ".join(parts)


class ImageScale(BrowserView):
    """view used for rendering image scales"""

    # Grant full access to this view even if the object being viewed is
    # protected
    # (it's okay because we explicitly validate access to the image attribute
    # when we retrieve it)
    __roles__ = ("Anonymous",)
    __allow_access_to_unprotected_subobjects__ = 1
    data = None

    # You can control which mimetypes may be shown inline
    # and which must always be downloaded, for security reasons.
    # Make the configuration available on the class.
    # Then subclasses can override this.
    allowed_inline_mimetypes = ALLOWED_INLINE_MIMETYPES
    disallowed_inline_mimetypes = DISALLOWED_INLINE_MIMETYPES
    use_denylist = USE_DENYLIST

    def __init__(self, context, request, **info):
        self.context = context
        self.request = request
        self.__dict__.update(**info)
        if self.data is None:
            self.data = getattr(self.context, self.fieldname)

        url = self.context.absolute_url()
        extension = self.data.contentType.split("/")[-1].lower()
        if self.data.contentType == "image/svg+xml":
            extension = "svg"
        if "uid" in info:
            name = info["uid"]
        else:
            name = info["fieldname"]
        self.__name__ = f"{name}.{extension}"
        self.url = f"{url}/@@images/{self.__name__}"
        self.srcset = info.get("srcset", [])

    def absolute_url(self):
        return self.url

    def srcset_attribute(self):
        _srcset_attr = []
        extension = self.data.contentType.split("/")[-1].lower()
        for scale in self.srcset:
            _srcset_attr.append(
                "{}/@@images/{}.{} {}x".format(
                    self.context.absolute_url(), scale["uid"], extension, scale["scale"]
                )
            )
        srcset_attr = ", ".join(_srcset_attr)
        return srcset_attr

    @property
    def title(self):
        """Get the title from the context.

        Let's not fail when we cannot find a title.
        """
        try:
            # Most Plone content items.
            return self.context.Title()
        except AttributeError:
            pass
        try:
            # Can work on a tile and most other things.
            return self.context.title
        except AttributeError:
            pass
        return ""

    def tag(
        self,
        height=_marker,
        width=_marker,
        alt=_marker,
        css_class=None,
        title=_marker,
        **kwargs,
    ):
        """Create a tag including scale"""
        if height is _marker:
            height = getattr(self, "height", self.data._height)
        if width is _marker:
            width = getattr(self, "width", self.data._width)

        if alt is _marker:
            alt = self.title
        if title is _marker:
            title = self.title

        values = [
            ("src", self.url),
            ("alt", alt),
            ("title", title),
            ("height", height),
            ("width", width),
            ("class", css_class),
        ]

        srcset_attr = self.srcset_attribute()
        if srcset_attr:
            values.append(("srcset", srcset_attr))

        values.extend(kwargs.items())
        return _image_tag_from_values(*values)

    def validate_access(self):
        fieldname = getattr(self.data, "fieldname", getattr(self, "fieldname", None))
        guarded_getattr(self.context, fieldname)

    def _should_force_download(self):
        # If this returns True, the caller should call set_headers with a filename.
        if not hasattr(self.data, "contentType"):
            return
        mimetype = extract_media_type(self.data.contentType)
        if self.use_denylist:
            # We explicitly deny a few mimetypes, and allow the rest.
            return mimetype in self.disallowed_inline_mimetypes
        # Use the allowlist.
        # We only explicitly allow a few mimetypes, and deny the rest.
        return mimetype not in self.allowed_inline_mimetypes

    def set_headers(self, response=None):
        # set headers for the image
        image = self.data
        if response is None:
            response = self.request.response
        filename = None
        if self._should_force_download():
            # We MUST pass a filename, even a dummy one if needed.
            filename = getattr(image, "filename", getattr(self, "filename", None))
            if filename is None:
                filename = getattr(image, "fieldname", getattr(self, "fieldname", None))
                if filename is None:
                    filename = "image.ext"
        set_headers(image, response, filename=filename)

    def index_html(self):
        """download the image"""
        self.validate_access()
        self.set_headers()
        return stream_data(self.data)

    def manage_DAVget(self):
        """Get scale via webdav."""
        return self.manage_FTPget()

    def manage_FTPget(self):
        """Get scale via ftp."""
        return self.index_html()

    def __call__(self):
        # avoid the need to prefix with nocall: in TAL
        return self

    def HEAD(self, REQUEST, RESPONSE=None):
        """Obtain metainformation about the image implied by the request
        without transfer of the image itself
        """
        self.validate_access()
        self.set_headers(response=REQUEST.response)
        return ""

    HEAD.__roles__ = ("Anonymous",)


@implementer(ITraversable)
class ImmutableTraverser:
    def __init__(self, scale):
        self.scale = scale

    def traverse(self, name, furtherPath):
        if furtherPath:
            raise TraversalError("Do not know how to handle further path")
        else:
            if self.scale:
                return self.scale.tag()
            else:
                raise TraversalError(name)


@implementer(IImageScaleFactory)
class DefaultImageScalingFactory:
    def __init__(self, context):
        self.context = context
        # fieldname will be set for real in the __call__ method.
        self.fieldname = None

    def get_original_value(self, fieldname=None):
        """Get the image value.

        In most cases this will be a NamedBlobImage field.
        """
        fieldname = fieldname or self.fieldname
        if fieldname is not None:
            return getattr(self.context, fieldname, None)
        try:
            primary = IPrimaryFieldInfo(self.context, None)
        except TypeError:
            return
        if primary is None:
            return
        self.fieldname = primary.fieldname
        return primary.value

    def get_raw_data(self, orig_value):
        """Get the raw image data.

        The result may be an open file, in which case it is the responsibility
        of the caller to close it.  Or it may be a string.
        """
        orig_data = None
        try:
            orig_data = orig_value.open()
        except AttributeError:
            orig_data = getattr(aq_base(orig_value), "data", orig_value)
        if not orig_data:
            return
        # Handle cases where large image data is stored in FileChunks instead
        # of plain string
        if isinstance(orig_data, tuple(FILECHUNK_CLASSES)):
            # Convert data to 8-bit string
            # (FileChunk does not provide read() access)
            orig_data = str(orig_data)
        return orig_data

    def url(self):
        # url of the context
        return self.context.absolute_url()

    def get_quality(self):
        """Get scaled image quality setting from imaging control panel."""
        getScaledImageQuality = queryUtility(IScaledImageQuality)
        if getScaledImageQuality is None:
            return None
        return getScaledImageQuality()

    def update_parameters(self, **parameters):
        # If quality wasn't in the parameters, try the site's default scaling
        # quality if it exists.
        if "quality" not in parameters:
            quality = self.get_quality()
            if quality:
                parameters["quality"] = quality
        return parameters

    def create_scale(self, data, mode, height, width, **parameters):
        if "direction" in parameters:
            warnings.warn(
                "The 'direction' option is deprecated, use 'mode' instead.",
                DeprecationWarning,
            )
            mode = parameters.pop("direction")
        return scaleImage(data, mode=mode, height=height, width=width, **parameters)

    def handle_image(self, orig_value, orig_data, mode, height, width, **parameters):
        """Return a scaled image, its mimetype format, and width and height."""
        if getattr(orig_value, "contentType", "") == "image/svg+xml":
            # No need to scale, we can simply use the original data,
            # but report a different width and height.
            if isinstance(orig_data, (str)):
                orig_data = safe_encode(orig_data)
            if isinstance(orig_data, (bytes)):
                orig_data = BytesIO(orig_data)
            result = orig_data.read(), "svg+xml", (width, height)
            return result
        try:
            result = self.create_scale(
                orig_data, mode=mode, height=height, width=width, **parameters
            )
        except (ConflictError, KeyboardInterrupt):
            raise
        except Exception:
            logger.exception(
                'Could not scale "{!r}" of {!r}'.format(
                    orig_value,
                    self.url(),
                ),
            )
            return
        return result

    def __call__(
        self,
        fieldname=None,
        mode="scale",
        height=None,
        width=None,
        scale=None,
        **parameters,
    ):

        """Factory for image scales`.

        Note: the 'scale' keyword argument is ignored.
        You should pass a height and width.
        """
        # Save self.fieldname for use in self.get_original_value
        # and other methods where we do not pass the fieldname explicitly.
        self.fieldname = fieldname
        orig_value = self.get_original_value()
        if orig_value is None:
            return
        want_original = height is None and width is None
        if not want_original:
            if "direction" in parameters:
                warnings.warn(
                    "The 'direction' option is deprecated, use 'mode' instead.",
                    DeprecationWarning,
                )
                # We must get rid of this duplicate parameter, otherwise it ends up in
                # hashes and it negates the next condition.
                mode = parameters.pop("direction")
            if (
                not parameters
                and height
                and width
                and height == getattr(orig_value, "_height", None)
                and width == getattr(orig_value, "_width", None)
            ):
                want_original = True
        if want_original:
            # No special wishes, and the original image already has the
            # requested height and width.  Return the original.
            dummy, format_ = orig_value.contentType.split("/", 1)
            return orig_value, format_, (orig_value._width, orig_value._height)

        orig_data = self.get_raw_data(orig_value)
        if not orig_data:
            return

        parameters = self.update_parameters(**parameters)
        if "modified" in parameters:
            del parameters["modified"]
        try:
            result = self.handle_image(
                orig_value, orig_data, mode, height, width, **parameters
            )
        finally:
            # Make sure the file is closed to avoid error:
            # ZODB-5.5.1-py3.7.egg/ZODB/blob.py:339: ResourceWarning:
            # unclosed file <_io.FileIO ... mode='rb' closefd=True>
            if isinstance(orig_data, BlobFile):
                orig_data.close()
        if result is None:
            return

        # Note: the format may differ from the original.
        # For example a TIFF may have been turned into a PNG.
        data, format_, dimensions = result
        mimetype = f"image/{format_.lower()}"
        value = orig_value.__class__(
            data,
            contentType=mimetype,
            filename=orig_value.filename,
        )
        value.fieldname = self.fieldname

        return value, format_, dimensions


@implementer(ITraversable, IBrowserPublisher)
class ImageScaling(BrowserView):
    """view used for generating (and storing) image scales"""

    # Ignore some stacks to help with accessing via webdav, otherwise you get a
    # 404 NotFound error.
    _ignored_stacks = ("manage_DAVget", "manage_FTPget")
    _scale_view_class = ImageScale

    def publishTraverse(self, request, name):
        """used for traversal via publisher, i.e. when using as a url"""
        stack = request.get("TraversalRequestNameStack")
        image = None
        if stack and stack[-1] not in self._ignored_stacks:
            # field and scale name were given...
            scale = stack.pop()
            image = self.scale(name, scale)  # this is an aq-wrapped scale_view
            if image:
                return image
        elif "-" in name:
            # we got a uid...
            if "." in name:
                name, ext = name.rsplit(".", 1)
            storage = getMultiAdapter((self.context, None), IImageScaleStorage)
            info = storage.get_or_generate(name)
            if info is None:
                raise NotFound(self, name, self.request)
            scale_view = self._scale_view_class(self.context, self.request, **info)
            alsoProvides(scale_view, IStableImageScale)
            return scale_view
        else:
            # otherwise `name` must refer to a field...
            if "." in name:
                name, ext = name.rsplit(".", 1)
            value = self.get_orig_image(name)
            scale_view = self._scale_view_class(
                self.context,
                self.request,
                data=value,
                fieldname=name,
            )
            return scale_view
        raise NotFound(self, name, self.request)

    def browserDefault(self, request):
        # There's nothing in the path after /@@images
        raise BadRequest("Missing image scale path")

    def traverse(self, name, furtherPath):
        """used for path traversal, i.e. in zope page templates"""
        # validate access
        value = self.guarded_orig_image(name)
        if not furtherPath:
            image = self._scale_view_class(
                self.context,
                self.request,
                data=value,
                fieldname=name,
            )
        else:
            return ImmutableTraverser(self.scale(name, furtherPath[-1], pre=True))

        if image is not None:
            return image.tag()
        raise TraversalError(self, name)

    _sizes = None

    @deprecate("use property available_sizes instead")
    def getAvailableSizes(self, fieldname=None):
        if fieldname:
            warnings.warn(
                "fieldname was passed to deprecated getAvailableSizes, but "
                "will be ignored.",
                DeprecationWarning,
            )
        return self.available_sizes

    @property
    def available_sizes(self):
        # fieldname is ignored by default
        if self._sizes is None:
            sizes_util = queryUtility(IAvailableSizes)
            if sizes_util is None:
                self._sizes = {}
            else:
                self._sizes = sizes_util() or {}
        return self._sizes

    @available_sizes.setter
    def available_sizes(self, value):
        self._sizes = value

    def getImageSize(self, fieldname=None):
        if fieldname is not None:
            try:
                value = self.guarded_orig_image(fieldname)
            except Unauthorized:
                # This is a corner case that can be seen in some tests,
                # at least plone.app.caching and plone.formwidget.namedfile.
                # When it is *really* unauthorized to get this image,
                # it will go wrong somewhere else.
                value = None
            if value is None:
                return (0, 0)
            return value.getImageSize()
        value = IPrimaryFieldInfo(self.context).value
        return value.getImageSize()

    def guarded_orig_image(self, fieldname):
        # Note: you must not call this from publishTraverse.
        # No authentication has taken place there yet, so everyone is still anonymous.
        return guarded_getattr(self.context, fieldname, None)

    def get_orig_image(self, fieldname):
        # Get the image without doing permission checks.
        # Use guarded_orig_image instead of you want permission checks.
        return getattr(self.context, fieldname, None)

    @deprecate("use getHighPixelDensityScales instead")
    def getRetinaScales(self):
        return getHighPixelDensityScales()

    def getHighPixelDensityScales(self):
        return getHighPixelDensityScales()

    def modified(self, fieldname=None):
        """Provide a callable to return the modification time of content
        items, so stored image scales can be invalidated.
        """
        context = aq_base(self.context)
        if fieldname is not None:
            field = getattr(context, fieldname, None)
            modified = getattr(field, "modified", None)
            date = DateTime(modified or context._p_mtime)
        else:
            date = DateTime(context._p_mtime)
        return date.millis()

    def scale(
        self,
        fieldname=None,
        scale=None,
        height=None,
        width=None,
        mode="scale",
        pre=False,
        include_srcset=None,
        **parameters,
    ):
        if fieldname is None:
            try:
                primary = IPrimaryFieldInfo(self.context, None)
            except TypeError:
                return
            if primary is None:
                return  # 404
            fieldname = primary.fieldname
        if scale is not None:
            if width is not None or height is not None:
                logger.warning(
                    "A scale name and width/height are given. Those are "
                    "mutually exclusive: solved by ignoring width/height and "
                    "taking name",
                )
            available = self.available_sizes
            if scale not in available:
                return None  # 404
            width, height = available[scale]
        storage = getMultiAdapter(
            (self.context, functools.partial(self.modified, fieldname)),
            IImageScaleStorage,
        )
        if pre:
            scale_method = storage.pre_scale
        else:
            scale_method = storage.scale
        info = scale_method(
            fieldname=fieldname,
            height=height,
            width=width,
            mode=mode,
            scale=scale,
            **parameters,
        )
        if info is None:
            return  # 404

        # Do we want to include srcset info for HiDPI?
        # If there is no explicit True/False given, we look at the value of 'pre'.
        # When 'pre' is False, the visitor is requesting a scale via a url,
        # so we only want a single image and not any fancy extras.
        if include_srcset is None and pre:
            include_srcset = True
        if include_srcset:
            if "srcset" not in info:
                info["srcset"] = self.calculate_srcset(
                    fieldname=fieldname,
                    height=height,
                    width=width,
                    mode=mode,
                    scale=scale,
                    storage=storage,
                    **parameters,
                )
        if "fieldname" not in info:
            info["fieldname"] = fieldname
        scale_view = self._scale_view_class(self.context, self.request, **info)
        return scale_view

    def calculate_srcset(
        self,
        fieldname=None,
        scale=None,
        height=None,
        width=None,
        mode="scale",
        storage=None,
        **parameters,
    ):
        srcset = []
        if storage is None:
            return srcset
        (orig_width, orig_height) = self.getImageSize(fieldname)
        for hdScale in self.getHighPixelDensityScales():
            # Don't create retina scales larger than the source image.
            # We only care about the width, because height might be 65536.
            if width and orig_width and orig_width < width * hdScale["scale"]:
                continue
            parameters["quality"] = hdScale["quality"]
            scale_src = storage.pre_scale(
                fieldname=fieldname,
                height=height * hdScale["scale"] if height else height,
                width=width * hdScale["scale"] if width else width,
                mode=mode,
                **parameters,
            )
            if scale_src is None:
                continue
            scale_src["scale"] = hdScale["scale"]
            srcset.append(scale_src)
        return srcset

    def tag(
        self,
        fieldname=None,
        scale=None,
        height=None,
        width=None,
        mode="scale",
        **kwargs,
    ):
        scale = self.scale(fieldname, scale, height, width, mode, pre=True)
        return scale.tag(**kwargs) if scale else None

    def picture(
        self,
        fieldname=None,
        picture_variant=None,
        alt=None,
        css_class=None,
        title=_marker,
        **kwargs,
    ):
        img2picturetag = Img2PictureTag()
        picture_variant_config = get_picture_variants().get(picture_variant)
        if not picture_variant_config:
            logger.warning(
                "Could not find the given picture_variant %s, "
                "creating ordinary img tag instead!",
                picture_variant,
            )
            if picture_variant in self.available_sizes:
                # We have a bit of luck: we have a scale with the same name
                # as the picture variant.
                scale = picture_variant
            else:
                scale = None
            return self.tag(
                fieldname=fieldname,
                scale=scale,
                alt=alt,
                css_class=css_class,
                title=title,
                **kwargs,
            )

        sourceset = picture_variant_config.get("sourceset")
        scale = self.scale(fieldname, sourceset[-1].get("scale"), pre=True)
        attributes = {}
        attributes["class"] = css_class and [css_class] or []
        if not attributes["class"]:
            del attributes["class"]
        attributes["src"] = scale.url
        attributes["width"] = scale.width
        attributes["height"] = scale.height
        if title is _marker:
            attributes["title"] = self.context.Title()
        elif title:
            attributes["title"] = title
        if alt:
            attributes["alt"] = alt
        return img2picturetag.create_picture_tag(
            sourceset,
            attributes,
            resolve_urls=True,
            uid=scale.context.UID(),
            fieldname=fieldname,
        ).prettify()


class NavigationRootScaling(ImageScaling):
    @lazy_property
    def _supports_image_scales_metadata(self):
        # Do we have the image_scales in the portal_catalog?
        # Expected to be False on Plone 5.2, True on Plone 6.0.
        catalog = getToolByName(self.context, "portal_catalog")
        return "image_scales" in catalog._catalog.schema

    @lazy_property
    def _supports_hidpi(self):
        # Do we have any "old-style" high DPI scales (2x/3x)?
        return bool(self.getHighPixelDensityScales())

    def _scale_cachekey(method, self, brain, fieldname, **kwargs):
        return (
            self.context.absolute_url(),
            brain.UID,
            brain.modified,
            fieldname,
            kwargs,
        )

    @ram.cache(_scale_cachekey)
    def tag(self, brain, fieldname, **kwargs):
        if self._supports_image_scales_metadata:
            tag = self._tag_from_brain_image_scales(brain, fieldname, **kwargs)
            if tag:
                return tag
        obj = brain.getObject()
        images = obj.restrictedTraverse("@@images")
        tag = images.tag(fieldname, **kwargs)
        return tag

    def _tag_from_brain_image_scales(
        self,
        brain,
        fieldname,
        scale=None,
        alt=_marker,
        css_class=None,
        title=_marker,
        **kwargs,
    ):
        """Try to get a tag from the image_scales metadata.

        If we have any non-standard keyword arguments, we cannot use this method.
        Especially you cannot set a mode: we must use the default "scale" mode.

        Also, no old-style hidpi srcsets are included.  If the site has enabled this,
        we return nothing: this information is not (easily) available in the brain.
        """
        if self._supports_hidpi:
            return
        if not (brain and fieldname and scale):
            return
        if kwargs:
            # too many keyword arguments
            return
        if not getattr(brain, "image_scales", None):
            # no images here at all
            return
        if fieldname not in brain.image_scales:
            return
        try:
            # Note: per field we get a list with dicts.
            # For normal image fields this will always be one dict.
            # When the field is a RelationList pointing to images,
            # it might contain more.  That does not sound like something
            # we can support out of the box, so we always use the first one.
            # This probably makes the most sense in that corner case as well.
            data = brain.image_scales[fieldname][0]["scales"][scale]
        except (KeyError, IndexError):
            return

        # data has download, height and width
        if title is _marker:
            title = brain.Title
            if callable(title):
                # Brain may be a CatalogContentListingObject.
                title = title()
        if alt is _marker:
            alt = title

        # common case is a local path to "@@images/img-scale"
        # but it might be a custom url starting with "http"
        src = (
            data["download"]
            if data["download"].startswith("http")
            else f"{brain.getURL()}/{data['download']}"
        )

        values = [
            ("src", src),
            ("alt", alt),
            ("title", title),
            ("height", data["height"]),
            ("width", data["width"]),
        ]
        if css_class:
            values.append(("class", css_class))
        return _image_tag_from_values(*values)


def _scale_sort_key(item):
    key, value = item
    try:
        fieldname, width, uid = key.split("-")
        width = int(width)
    except (ValueError, IndexError, TypeError):
        return (key,)
    return (fieldname, width, uid)


class ImagesTest(BrowserView):
    """View for Editors to check how images look and what scales are stored."""

    @property
    def storage(self):
        return getMultiAdapter((self.context, None), IImageScaleStorage)

    def stored_scales(self):
        return sorted(self.storage.items(), key=_scale_sort_key)

    def clear(self):
        """Clear the scales."""
        PostOnly(self.request)
        self.storage.clear()
        url = self.context.absolute_url()
        logger.info("Scale storage cleared for %s", url)
        self.request.response.redirect(f"{url}/@@images-test")
        return "cleared"



import unittest

from plone.namedfile.file import NamedImage
from plone.namedfile.tests import getFile
from plone.namedfile.utils import get_contenttype
from plone.namedfile.utils.svg_utils import dimension_int
from plone.namedfile.utils.svg_utils import process_svg


class TestSvg(unittest.TestCase):
    def test_get_contenttype(self):
        self.assertEqual(
            get_contenttype(
                NamedImage(getFile("image.svg"), contentType="image/svg+xml")
            ),
            "image/svg+xml",
        )

    def test_process_svg(self):

        content_type, width, height = process_svg(getFile("image.svg"))
        self.assertEqual(content_type, "image/svg+xml")
        self.assertEqual(width, 158)
        self.assertEqual(height, 40)

    def test_process_svg__indicate_header_truncation(self):
        """ Check that we can detect SVG files where the file header was
         larger than the requested first bytes. process_svg() should
         return -1 as dimensions to indicate the truncation."""

        truncated_data = getFile("image_large_header.svg", length=1024)
        content_type, width, height = process_svg(truncated_data)
        self.assertEqual(content_type, "image/svg+xml")
        self.assertEqual(width, -1)
        self.assertEqual(height, -1)

    def test_process_svg__can_handle_large_header(self):

        data = getFile("image_large_header.svg")
        content_type, width, height = process_svg(data)
        self.assertEqual(content_type, "image/svg+xml")
        self.assertEqual(width, 1041)
        self.assertEqual(height, 751)

    def test_dimension_int(self):

        self.assertEqual(dimension_int("auto"), 0)
        self.assertEqual(dimension_int("0"), 0)
        self.assertEqual(dimension_int("25"), 25)
        self.assertEqual(dimension_int("1024px"), 1024)
        self.assertEqual(dimension_int("123.0025px"), 123)
        self.assertEqual(dimension_int("123.625pt"), 123)
        self.assertEqual(dimension_int("50%"), 50)
        self.assertEqual(dimension_int(42), 42)
        self.assertEqual(dimension_int(6.25), 6)
        self.assertEqual(dimension_int({}), 0)
        self.assertEqual(dimension_int([]), 0)
        self.assertEqual(dimension_int(getFile("image.svg")), 0)



from plone.namedfile.file import NamedBlobFile  # noqa
from plone.namedfile.file import NamedBlobImage  # noqa
from plone.namedfile.file import NamedFile  # noqa
from plone.namedfile.file import NamedImage  # noqa



from plone.namedfile import field
from plone.supermodel.exportimport import ObjectHandler


class FileFieldHandler(ObjectHandler):
    filteredAttributes = ObjectHandler.filteredAttributes.copy()
    filteredAttributes.update({"default": "rw", "missing_value": "rw", "schema": "rw"})


NamedFileHandler = FileFieldHandler(field.NamedFile)
NamedImageHandler = FileFieldHandler(field.NamedImage)

NamedBlobFileHandler = FileFieldHandler(field.NamedBlobFile)
NamedBlobImageHandler = FileFieldHandler(field.NamedBlobImage)



from DateTime import DateTime
from io import BytesIO
from OFS.SimpleItem import SimpleItem
from plone.app.testing import TEST_USER_NAME
from plone.app.testing import TEST_USER_PASSWORD
from plone.namedfile.field import NamedImage as NamedImageField
from plone.namedfile.file import NamedImage
from plone.namedfile.interfaces import IImageScaleTraversable
from plone.namedfile.scaling import ImageScaling
from plone.namedfile.testing import PLONE_NAMEDFILE_FUNCTIONAL_TESTING
from plone.namedfile.tests import getFile
from plone.testing.zope import Browser
from zExceptions import BadRequest
from zope.annotation import IAttributeAnnotatable
from zope.interface import implementer

import PIL
import transaction
import unittest


class IHasImage(IImageScaleTraversable):
    image = NamedImageField()


def assertImage(testcase, data, format_, size):
    image = PIL.Image.open(BytesIO(data))
    testcase.assertEqual(image.format, format_)
    testcase.assertEqual(image.size, size)


@implementer(IAttributeAnnotatable, IHasImage)
class DummyContent(SimpleItem):
    image = None
    modified = DateTime
    id = __name__ = "item"
    title = "foo"

    def Title(self):
        return self.title


class ImagePublisherTests(unittest.TestCase):

    layer = PLONE_NAMEDFILE_FUNCTIONAL_TESTING

    def setUp(self):
        data = getFile("image.png")
        item = DummyContent()
        item.image = NamedImage(data, "image/png", "image.png")
        self.layer["app"]._setOb("item", item)
        self.item = self.layer["app"].item
        self.view = self.item.unrestrictedTraverse("@@images")
        self._orig_sizes = ImageScaling._sizes

        self.browser = Browser(self.layer["app"])
        self.browser.handleErrors = False
        self.browser.addHeader("Referer", self.layer["app"].absolute_url())

    def tearDown(self):
        ImageScaling._sizes = self._orig_sizes

    def testPublishScaleViaUID(self):
        scale = self.view.scale("image", width=64, height=64)
        transaction.commit()
        # make sure the referenced image scale is available
        self.browser.open(scale.url)
        self.assertEqual("image/png", self.browser.headers["content-type"])
        assertImage(self, self.browser.contents, "PNG", (64, 64))

    def testPublishWebDavScaleViaUID(self):
        scale = self.view.scale("image", width=64, height=64)
        transaction.commit()
        # make sure the referenced image scale is available
        self.browser.open(scale.url + "/manage_DAVget")
        self.assertEqual("image/png", self.browser.headers["content-type"])
        assertImage(self, self.browser.contents, "PNG", (64, 64))

    def testPublishFTPScaleViaUID(self):
        scale = self.view.scale("image", width=64, height=64)
        transaction.commit()
        # make sure the referenced image scale is available
        self.browser.open(scale.url + "/manage_FTPget")
        self.assertIn("200", self.browser.headers["status"])
        # Same remark as in testPublishWebDavScaleViaUID is valid here.
        self.assertEqual("image/png", self.browser.headers["content-type"])
        assertImage(self, self.browser.contents, "PNG", (64, 64))

    def testHeadRequestMethod(self):
        scale = self.view.scale("image", width=64, height=64)
        transaction.commit()
        # make sure the referenced image scale is available
        url = str(scale.url)
        self.browser.open(url)
        GET_length = len(self.browser.contents)

        self.browser = Browser(self.layer["app"])
        self.browser.handleErrors = False
        self.browser.addHeader("Referer", self.layer["app"].absolute_url())

        def make_head_request(args):
            return self.browser.testapp.head(url, **args)

        self.browser._processRequest(url, make_head_request)

        self.assertEqual("image/png", self.browser.headers["content-type"])
        self.assertEqual(self.browser.headers["Content-Length"], str(GET_length))
        self.assertEqual(self.browser.contents, b"")

    def testPublishThumbViaUID(self):
        ImageScaling._sizes = {"thumb": (128, 128)}
        scale = self.view.scale("image", "thumb")
        transaction.commit()
        # make sure the referenced image scale is available
        self.browser.open(scale.url)
        self.assertEqual("image/png", self.browser.headers["content-type"])
        assertImage(self, self.browser.contents, "PNG", (128, 128))

    def testPublishCustomSizeViaUID(self):
        # set custom image sizes
        ImageScaling._sizes = {"foo": (23, 23)}
        scale = self.view.scale("image", "foo")
        transaction.commit()
        # make sure the referenced image scale is available
        self.browser.open(scale.url)
        self.assertEqual("image/png", self.browser.headers["content-type"])
        assertImage(self, self.browser.contents, "PNG", (23, 23))

    def testPublishThumbViaName(self):
        ImageScaling._sizes = {"thumb": (128, 128)}
        transaction.commit()

        # make sure traversing works as is and with scaling
        # first the field without a scale name
        self.browser.open(self.layer["app"].absolute_url() + "/item/@@images/image")
        self.assertEqual("image/png", self.browser.headers["content-type"])
        self.assertEqual(self.browser.contents, getFile("image.png"))

        # and last a scaled version
        self.browser.open(
            self.layer["app"].absolute_url() + "/item/@@images/image/thumb"
        )
        self.assertEqual("image/png", self.browser.headers["content-type"])
        assertImage(self, self.browser.contents, "PNG", (128, 128))

    def testPublishCustomSizeViaName(self):
        # set custom image sizes
        ImageScaling._sizes = {"foo": (23, 23)}
        transaction.commit()
        # make sure traversing works as expected
        self.browser.open(self.layer["app"].absolute_url() + "/item/@@images/image/foo")
        assertImage(self, self.browser.contents, "PNG", (23, 23))

    def testPublishScaleWithInvalidUID(self):
        scale = self.view.scale("image", width=64, height=64)
        transaction.commit()
        # change the url so it's invalid...
        from zExceptions import NotFound

        with self.assertRaises(NotFound):
            self.browser.open(scale.url.replace(".png", "x.png"))

    def testPublishScaleWithInvalidScale(self):
        scale = self.view.scale("image", "no-such-scale")
        transaction.commit()
        self.assertEqual(scale, None)

    def test_getAvailableSizesWithInvalidScaleMethod(self):
        self.assertEqual(self.view.getAvailableSizes("no-such-scale"), {})

    def test_getAvailableSizesWithInvalidScaleProperty(self):
        self.assertEqual(self.view.available_sizes, {})

    def test_getImageSizeWithInvalidScale(self):
        self.assertEqual(self.view.getImageSize("no-such-scale"), (0, 0))

    def testGuardedAccess(self):
        # make sure it's not possible to access scales of forbidden images
        self.item.__allow_access_to_unprotected_subobjects__ = 0
        ImageScaling._sizes = {"foo": (23, 23)}
        transaction.commit()
        self.browser.addHeader(
            "Authorization",
            f"Basic {TEST_USER_NAME:s}:{TEST_USER_PASSWORD:s}",
        )
        from zExceptions import Unauthorized

        with self.assertRaises(Unauthorized):
            self.browser.open(
                self.layer["app"].absolute_url() + "/item/@@images/image/foo"
            )
        self.item.__allow_access_to_unprotected_subobjects__ = 1

    def testSVGPublishThumbViaName(self):
        ImageScaling._sizes = {"thumb": (128, 128)}

        data = getFile("image.svg")
        svg = DummyContent()
        svg.image = NamedImage(data, "image/svg+xml", "image.svg")
        self.layer["app"]._setOb("svg", svg)
        svg = self.layer["app"].svg

        transaction.commit()

        # make sure traversing works as is and with scaling
        # first the field without a scale name
        self.browser.open(self.layer["app"].absolute_url() + "/svg/@@images/image")

        self.assertEqual("image/svg+xml", self.browser.headers["content-type"])
        self.assertEqual(self.browser.contents, getFile("image.svg"))

        # and last a scaled version
        self.browser.open(
            self.layer["app"].absolute_url() + "/svg/@@images/image/thumb"
        )
        self.assertEqual("image/svg+xml", self.browser.headers["content-type"])
        self.assertEqual(self.browser.contents, data)

    def testImagesViewWithNoSubpath(self):
        transaction.commit()
        with self.assertRaises(BadRequest):
            self.browser.open(self.layer["app"].absolute_url() + "/item/@@images")


def test_suite():
    from unittest import defaultTestLoader

    return defaultTestLoader.loadTestsFromName(__name__)



from setuptools import find_packages
from setuptools import setup

import os


version = "6.4.0"

description = "File types and fields for images, files and blob files with filenames"
long_description = "\n\n".join(
    [
        open("README.rst").read(),
        open("CHANGES.rst").read(),
        open(os.path.join("plone", "namedfile", "usage.rst")).read(),
    ]
)


setup(
    name="plone.namedfile",
    version=version,
    description=description,
    long_description=long_description,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Framework :: Plone",
        "Framework :: Plone :: 5.2",
        "Framework :: Plone :: 6.0",
        "Framework :: Plone :: Core",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "License :: OSI Approved :: BSD License",
    ],
    keywords="plone named file image blob",
    author="Laurence Rowe, Martin Aspeli",
    author_email="plone-developers@lists.sourceforge.net",
    url="https://pypi.org/project/plone.namedfile",
    license="BSD",
    packages=find_packages(),
    namespace_packages=["plone"],
    include_package_data=True,
    zip_safe=False,
    python_requires=">=3.7",
    install_requires=[
        "persistent",
        "piexif",
        "plone.app.uuid>=2.2.0",
        "plone.rfc822>=2.0.0",
        "plone.scale[storage] >=3.0",
        "plone.schemaeditor",
        "plone.supermodel",
        "setuptools",
        "zope.browserpage",
        "zope.cachedescriptors",
        "zope.component",
        "zope.copy",
        "zope.security",
        "zope.traversing",
    ],
    extras_require={
        "test": [
            "plone.app.testing",
            "lxml",
            "Pillow",
            "plone.testing[z2]",
        ],
    },
)



from logging import getLogger

import struct


log = getLogger(__name__)


def process_png(data):
    content_type = None
    w = -1
    h = -1
    size = len(data)
    # See PNG 2. Edition spec (http://www.w3.org/TR/PNG/)
    # Bytes 0-7 are below, 4-byte chunk length, then 'IHDR'
    # and finally the 4-byte width, height
    if (
        (size >= 24)
        and data.startswith(b"\211PNG\r\n\032\n")
        and (data[12:16] == b"IHDR")
    ):
        content_type = "image/png"
        w, h = struct.unpack(">LL", data[16:24])

    # Maybe this is for an older PNG version.
    elif (size >= 16) and data.startswith(b"\211PNG\r\n\032\n"):
        # Check to see if we have the right content type
        content_type = "image/png"
        w, h = struct.unpack(b">LL", data[8:16])

    width = int(w)
    height = int(h)
    return content_type, width, height



# This file was borrowed from z3c.blobfile and is licensed under the terms of
# the ZPL.

##############################################################################
#
# Copyright (c) 2008 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

from DateTime import DateTime
from plone.namedfile import storages
from plone.namedfile.file import NamedBlobFile
from plone.namedfile.file import NamedBlobImage
from plone.namedfile.interfaces import INamedBlobFile
from plone.namedfile.interfaces import INamedBlobImage
from plone.namedfile.interfaces import IStorage
from plone.namedfile.testing import PLONE_NAMEDFILE_FUNCTIONAL_TESTING
from plone.namedfile.testing import PLONE_NAMEDFILE_INTEGRATION_TESTING
from plone.namedfile.tests.test_image import zptlogo
from plone.namedfile.tests import MockNamedBlobImage
from zope.component import provideUtility
from zope.interface.verify import verifyClass

import struct
import time
import transaction
import unittest


def registerUtilities():
    provideUtility(storages.StringStorable(), IStorage, name="__builtin__.str")
    provideUtility(storages.UnicodeStorable(), IStorage, name="__builtin__.unicode")
    provideUtility(
        storages.FileChunkStorable(), IStorage, name="plone.namedfile.file.FileChunk"
    )
    provideUtility(storages.FileDescriptorStorable(), IStorage, name="__builtin__.file")


class TestImage(unittest.TestCase):

    layer = PLONE_NAMEDFILE_INTEGRATION_TESTING

    def setUp(self):
        registerUtilities()

    def _makeImage(self, *args, **kw):
        return NamedBlobImage(*args, **kw)

    def testEmpty(self):
        file = self._makeImage()
        self.assertEqual(file.contentType, "")
        self.assertEqual(file.data, b"")
        self.assertIsNotNone(file.modified)

    def testConstructor(self):
        file = self._makeImage(b"Data")
        self.assertEqual(file.contentType, "")
        self.assertEqual(file.data, b"Data")
        self.assertIsNotNone(file.modified)

    def testMutators(self):
        image = self._makeImage()

        image.contentType = "image/jpeg"
        self.assertEqual(image.contentType, "image/jpeg")

        image._setData(zptlogo)
        self.assertEqual(image.data, zptlogo)
        self.assertEqual(image.contentType, "image/gif")
        self.assertEqual(image.getImageSize(), (16, 16))

    def testModifiedTimeStamp(self):
        image = self._makeImage()
        old_timestamp = image.modified
        time.sleep(1/1000)  # make sure at least 1ms passes
        now = DateTime()
        self.assertGreater(now, DateTime(old_timestamp))
        image._setData(zptlogo)
        self.assertNotEqual(image.modified, old_timestamp)

    def testFallBackToDatabaseModifiedTimeStamp(self):
        dt = DateTime()
        image = MockNamedBlobImage()
        image._p_mtime = int(dt)
        image._modified = (dt + 1).millis()

        delattr(image, "_modified")
        self.assertEqual(image.modified, image._p_mtime)

    def testInterface(self):
        self.assertTrue(INamedBlobImage.implementedBy(NamedBlobImage))
        self.assertTrue(verifyClass(INamedBlobImage, NamedBlobImage))
        self.assertTrue(INamedBlobFile.implementedBy(NamedBlobImage))
        self.assertTrue(INamedBlobImage.implementedBy(NamedBlobImage))
        self.assertTrue(verifyClass(INamedBlobFile, NamedBlobImage))

    def testDataMutatorWithLargeJPGHeader(self):
        from plone.namedfile.file import IMAGE_INFO_BYTES

        bogus_header_length = struct.pack(">H", IMAGE_INFO_BYTES * 2)
        data = (
            b"\xff\xd8\xff\xe0"
            + bogus_header_length
            + b"\x00" * IMAGE_INFO_BYTES * 2
            + b"\xff\xc0\x00\x11\x08\x02\xa8\x04\x00"
        )
        image = self._makeImage()
        image._setData(data)
        self.assertEqual(image.getImageSize(), (1024, 680))

    def testDataMutatorWithLargeSVGHeader(self):
        from plone.namedfile.file import IMAGE_INFO_BYTES

        to_big_header_data = b'd' * (IMAGE_INFO_BYTES * 2)

        data = (
            b'<svg xmlns="http://www.w3.org/2000/svg" '
            b'width="1024px" '
            b'height="680px" '
            b'foobar="' + to_big_header_data + b'">'
            b'</svg>"'
        )
        image = self._makeImage()
        image._setData(data)
        self.assertEqual(image.getImageSize(), (1024, 680))
        self.assertGreater(len(to_big_header_data), IMAGE_INFO_BYTES)


class TestImageFunctional(unittest.TestCase):

    layer = PLONE_NAMEDFILE_FUNCTIONAL_TESTING

    def setUp(self):
        registerUtilities()

    def testCopyBlobs(self):
        from zope.copy import copy

        file = NamedBlobFile()
        file.data = "hello, world"
        image = NamedBlobImage()
        image.data = "some image bytes"
        transaction.commit()

        file_copy = copy(file)
        self.assertEqual(file_copy.data, file.data)

        image_copy = copy(image)
        self.assertEqual(image_copy.data, image.data)
